from collections import defaultdict


class EarlyStopping(object):

    def __init__(self, patience: int = 0, tolerance: float = 0, stop_all_folds: bool = True):
        self.patience = patience
        self.tolerance = tolerance
        self.stop_all_folds = stop_all_folds
        self.last_best_losses = defaultdict(lambda: float('inf'))
        self.counts = defaultdict(lambda: 0)
        self.call_counter = 0

    def __call__(self, fold, val_loss):
        self.call_counter += 1
        if val_loss - self.last_best_losses[fold] < self.tolerance:
            self.last_best_losses[fold] = val_loss
        else:
            self.counts[fold] += 1
            if self.counts[fold] > self.patience:
                if self.stop_all_folds:
                    raise StopIteration()
                else:
                    raise StopIteration(fold)



from unittest import TestCase

from sklearn.linear_model import Lasso
from sklearn.neural_network import MLPRegressor, MLPClassifier
from sklearn.datasets import make_regression, make_classification
from pandas_ml_common import np, pd, naive_splitter, stratified_random_splitter
from pandas_ml_utils.ml.model.base_model import AutoEncoderModel
from pandas_ml_utils_test.ml.model.test_abstract_model import TestAbstractModel
from pandas_ml_utils import SkModel, SkAutoEncoderModel, FeaturesAndLabels, ClassificationSummary, RegressionSummary, FittingParameter
from pandas_ml_utils_test.config import DF_NOTES


class TestSkModel(TestAbstractModel, TestCase):

    def test_linear_model(self):
        df = DF_NOTES.copy()

        with df.model() as m:
            fit = m.fit(
                SkModel(
                    Lasso(),
                    FeaturesAndLabels(
                        features=[
                            lambda df: df["variance"],
                            lambda df: (df["skewness"] / df["kurtosis"]).rename("engineered")
                        ],
                        labels=[
                            'authentic'
                        ]
                    )
                ),
                FittingParameter(naive_splitter())
            )

        print(fit)

        prediction = df.model.predict(fit.model)
        print(prediction)

        backtest = df.model.backtest(fit.model)
        self.assertLess(backtest.model.sk_model.coef_[0], 1e-5)

    def test_partial_fit_regression(self):
        data = make_regression(100, 2, 1)
        df = pd.DataFrame(data[0])
        df["label"] = data[1]

        with df.model() as m:
            fit_partial = m.fit(
                SkModel(
                    MLPRegressor(max_iter=1, random_state=42),
                    FeaturesAndLabels(features=[0, 1], labels=['label'])
                ),
                FittingParameter(
                    naive_splitter(0.3),
                    batch_size=10,
                    fold_epochs=10
                )
            )

        with df.model() as m:
            fit = m.fit(
                SkModel(
                    MLPRegressor(max_iter=10, random_state=42),
                    FeaturesAndLabels(features=[0, 1], labels=['label'])
                ),
                FittingParameter(naive_splitter(0.3))
            )

        self.assertAlmostEqual(df.model.predict(fit.model).iloc[0,-1], df.model.predict(fit_partial.model).iloc[0,-1], 4)

    def test_partial_fit_classification(self):
        data = make_classification(100, 2, 1, 0, n_clusters_per_class=1)
        df = pd.DataFrame(data[0])
        df["label"] = data[1]

        with df.model() as m:
            fit_partial = m.fit(
                SkModel(
                    MLPClassifier(max_iter=1, random_state=42),
                    FeaturesAndLabels(features=[0, 1], labels=['label']),
                    classes=np.unique(data[1])
                ),
                FittingParameter(
                    stratified_random_splitter(0.3),
                    batch_size=10,
                    fold_epochs=10,
                )
            )

        with df.model() as m:
            fit = m.fit(
                SkModel(
                    MLPClassifier(max_iter=10, random_state=42),
                    FeaturesAndLabels(features=[0, 1], labels=['label'])
                ),
                FittingParameter(stratified_random_splitter(0.3))
            )

        self.assertAlmostEqual(df.model.predict(fit.model).iloc[0,-1], df.model.predict(fit_partial.model).iloc[0,-1], 4)

    def provide_classification_model(self, features_and_labels):
        model = SkModel(
            MLPClassifier(activation='logistic', max_iter=1000, hidden_layer_sizes=(3,), alpha=0.001, solver='lbfgs', random_state=42),
            features_and_labels,
            summary_provider=ClassificationSummary
        )

        return model

    def provide_regression_model(self, features_and_labels):
        model = SkModel(
            MLPRegressor(1, learning_rate_init=0.01, solver='sgd', activation='identity', momentum=0, max_iter=1500, n_iter_no_change=500, nesterovs_momentum=False, shuffle=False, validation_fraction=0.0, random_state=42),
            features_and_labels,
            summary_provider=RegressionSummary
        )

        return model

    def provide_auto_encoder_model(self, features_and_labels) -> AutoEncoderModel:
        model = SkAutoEncoderModel(
            [2, 1], [2],
            features_and_labels,
            #learning_rate_init=0.0001,
            solver='lbfgs',
            validation_fraction=0,
            activation='logistic',
            momentum=0,
            max_iter=200,
            n_iter_no_change=500,
            nesterovs_momentum=False,
            shuffle=False,
            random_state=10
        )

        return model


import os
import unittest
from unittest import TestCase

from pandas_ml_common_test.notebook_runner import run_notebook
from pandas_ml_utils import Model
from pandas_ml_utils_test.config import DF_NOTES
import subprocess
import sys

PWD = os.path.dirname(os.path.abspath(__file__))
BIN = sys.executable
MODEL_FILE = os.path.join(PWD, 'test_ctx.model')
CLEAN_SCRIPT = os.path.join(PWD, "subprocess_load_module.py")


def test_model_import(*args):
    with subprocess.Popen([BIN, CLEAN_SCRIPT, *args], stdout=subprocess.PIPE) as p:
        return p.stdout.read().decode('utf-8')


class TestModelContext(TestCase):

    @unittest.skip("only used for debugging")
    def test_subprocess(self):
        out = test_model_import(MODEL_FILE)
        print(out)

        self.assertIn(MODEL_FILE, out)
        self.assertIn("MLPRegressor", out)
        self.assertIn("authentic", out)

    def test_save_and_load_model_from_context(self):
        """given a clean workspace"""
        notebook_file = os.path.join(PWD, f'{__file__}.ipynb')
        out_notebook_file = os.path.join(PWD, f'{__file__}.py_all_output.ipynb')

        if os.path.exists(MODEL_FILE): os.remove(MODEL_FILE)
        if os.path.exists(out_notebook_file): os.remove(out_notebook_file)

        """when executing the notebook with a model context"""
        nb, errors = run_notebook(notebook_file, PWD, kernel=os.getenv("TOX_KERNEL") or "python3")

        """then we have no errors and a saved model"""
        self.assertEqual(errors, [])
        self.assertTrue(os.path.exists(MODEL_FILE))

        """and when we load the saved model"""
        model = Model.load(MODEL_FILE)

        """then we can execute it"""
        self.assertEqual(2, len(DF_NOTES.model.predict(model, tail=2)))

        """and we can load the odel in a clean name space"""
        out = test_model_import(MODEL_FILE)
        print(out)

        self.assertIn(MODEL_FILE, out)
        self.assertIn("authentic", out)








from typing import Any, Callable

import pandas as pd

import pandas_ml_utils.html as html
from pandas_ml_common import Typing
from pandas_ml_common.utils.serialization_utils import plot_to_html_img
from pandas_ml_utils.ml.summary import Summary


class Fit(object):
    """
    After a model is fitted it gets embedded into this class along with some :class:`.Summary` statistics.
    In the case of `Fit` is displayed in a notebook the _repr_html of the Fit and Summary objects are called.
    """

    def __init__(self,
                 model: 'Model',
                 training_summary: Summary,
                 test_summary: Summary,
                 trails: Any = None,
                 **kwargs):
        self.model = model
        self.training_summary = training_summary
        self.test_summary = test_summary
        self._trails = trails
        self._kwargs = kwargs
        self._hide_loss_plot = False

    def plot_loss(self, figsize=(8, 6), **kwargs):
        return self.model.plot_loss(figsize, **kwargs)

    def values(self):
        """
        :return: returns the fitted model, a :class:`.Summary` on the training data, a :class:`.Summary` on the test data
        """
        return self.model, self.training_summary, self.test_summary

    def trails(self):
        """
        In case of hyper parameter optimization a trails object as used by `Hyperopt <https://github.com/hyperopt/hyperopt/wiki/FMin>`_
        is available.

        :return: Trails object
        """
        if self._trails is not None:
            return pd.DataFrame(self._trails.results)\
                     .drop("parameter", axis=1)\
                     .join(pd.DataFrame([r['parameter'] for r in self._trails.results]))
        else:
            return None

    def save_model(self, filename: str):
        """
        Save the fitted model.

        :param filename: filename
        :return: None
        """
        self.model.save(filename)

    def with_hidden_loss_plot(self):
        self._hide_loss_plot = True
        return self

    def with_summary(self, summary_provider: Callable[[Typing.PatchedDataFrame], Summary] = Summary, **kwargs):
        return Fit(self.model,
                   summary_provider(self.training_summary.df, **{**self._kwargs, **kwargs}),
                   summary_provider(self.test_summary.df, **{**self._kwargs, **kwargs}),
                   self._trails,
                   **self._kwargs)

    def __str__(self):
        summaries = f"train:\n{self.training_summary}\ntest:\n{self.test_summary}"
        backend = None
        try:
            import matplotlib
            backend = matplotlib.get_backend()
            matplotlib.use('module://drawilleplot')
            fig = self.model.plot_loss()
            return f"{fig}\n{summaries}"
        except:
            return summaries
        finally:
            if backend is not None:
                import matplotlib
                matplotlib.use(backend)

    def _repr_html_(self):
        from mako.template import Template
        from mako.lookup import TemplateLookup

        template = Template(filename=html.FIT_TEMPLATE, lookup=TemplateLookup(directories=['/']))
        return template.render(
            fit=self,
            loss_plot=plot_to_html_img(self.model.plot_loss, **self._kwargs) if not self._hide_loss_plot else None
        )


import logging
import os
from abc import abstractmethod
from collections import defaultdict
from copy import deepcopy
from functools import partial
from typing import Callable, Tuple, List

import dill as pickle
import numpy as np
import pandas as pd

from pandas_ml_common import Typing, Sampler, LazyInit
from pandas_ml_common.sampling.sampler import XYWeight
from pandas_ml_common.utils import merge_kwargs, call_callable_dynamic_args
from pandas_ml_utils.constants import PREDICTION_COLUMN_NAME
from pandas_ml_utils.ml.data.extraction import FeaturesAndLabels
from pandas_ml_utils.ml.fitting import FittingParameter
from pandas_ml_utils.ml.summary import Summary

_log = logging.getLogger(__name__)


class Model(object):
    """
    Represents a statistical or ML model and holds the necessary information how to interpret the columns of a
    pandas *DataFrame* ( :class:`.FeaturesAndLabels` ). Currently available implementations are:
     * SkitModel - provide any skit learn classifier or regressor
     * KerasModel - provide a function returning a compiled keras model
     * MultiModel - provide a model which will copied (and fitted) for each provided target
    """

    @staticmethod
    def load(filename: str):
        """
        Loads a previously saved model from disk. By default `dill <https://pypi.org/project/dill/>`_ is used to
        serialize / deserialize a model.

        :param filename: filename of the serialized model inclusive file extension
        :return: returns a deserialized model
        """
        with open(filename, 'rb') as file:
            model = pickle.load(file)
            if isinstance(model, Model):
                return model
            else:
                raise ValueError("Deserialized pickle was not a Model!")

    def __init__(self,
                 features_and_labels: FeaturesAndLabels,
                 summary_provider: Callable[[Typing.PatchedDataFrame], Summary] = Summary,
                 **kwargs):
        """
        All implementations of `Model` need to pass two arguments to `super().__init()__`.

        :param features_and_labels: the :class:`.FeaturesAndLabels` object defining all the features,
                                    feature engineerings and labels
        :param summary_provider: a summary provider in the most simple case just holds a `pd.DataFrame` containing all
                                 the labels and all the predictions and optionally loss and target values. Since
                                 constructors as callables as well it is usually enough tho just pass the type i.e.
                                 `summary_provider=BinaryClassificationSummary`
        :param kwargs:
        """
        self._features_and_labels = features_and_labels
        self._summary_provider = summary_provider
        self._history = defaultdict(dict)
        self._labels_columns = None
        self._feature_columns = None
        self._fit_meta_data: FittingParameter = None
        self.kwargs = kwargs

    @property
    def features_and_labels(self):
        return self._features_and_labels

    @property
    def summary_provider(self):
        return self._summary_provider

    def fit(self, x_y_weights: XYWeight, fitting_parameter: FittingParameter, verbose: int = 0, callbacks=None, **kwargs) -> Tuple[Typing.PatchedDataFrame, Typing.PatchedDataFrame]:
        self._fit_meta_data = fitting_parameter
        self.init_fit(**kwargs)
        processed_batches = 0

        sampler = self._sampler_with_callbacks(
            Sampler(
                x_y_weights,
                splitter=fitting_parameter.splitter,
                filter=fitting_parameter.filter,
                cross_validation=fitting_parameter.cross_validation,
                epochs=fitting_parameter.epochs,
                fold_epochs=fitting_parameter.fold_epochs,
                batch_size=fitting_parameter.batch_size
            ),
            verbose,
            callbacks
        )

        for batch in sampler.sample_for_training():
            self.fit_batch(batch.x, batch.y, batch.weight, batch.fold, **kwargs)
            processed_batches += 1

        if processed_batches <= 0:
            raise ValueError(f"Not enough data {[len(f) for f in sampler.frames[0]]}")

        training_data = sampler.get_in_sample_features()
        df_training_prediction = self.train_predict(training_data, **kwargs)

        test_data = sampler.get_out_of_sample_features()
        df_test_prediction = self.predict(test_data) if len(test_data) > 0 else pd.DataFrame({})

        return df_training_prediction, df_test_prediction

    def _sampler_with_callbacks(self, sampler: Sampler, verbose: int = 0, callbacks=None) -> Sampler:
        return sampler.with_callbacks(
            on_start=self._record_meta,
            on_fold=self.init_fold,
            after_fold_epoch=partial(self._record_loss, callbacks=callbacks, verbose=verbose),
            after_epoch=self.merge_folds,
            after_end=self.finish_learning
        )

    def _record_meta(self, epochs, batch_size, fold_epochs, cross_validation, features, labels: List[str]):
        partial_fit = any([size > 1 for size in [epochs, batch_size, fold_epochs] if size is not None])
        self._labels_columns = labels
        self._feature_columns = features

    def _record_loss(self, epoch, fold, fold_epoch, train_data: XYWeight, test_data: List[XYWeight], verbose, callbacks, loss_history_key=None):
        train_loss = self.calculate_loss(fold, train_data.x, train_data.y, train_data.weight)
        self._history["train", loss_history_key or fold][(epoch, fold_epoch)] = train_loss

        if len(test_data) > 0:
            test_loss = np.array([self.calculate_loss(fold, x, y, w) for x, y, w in test_data if len(x) > 0]).mean()
        else:
            test_loss = np.NaN
        self._history["test", loss_history_key or fold][(epoch, fold_epoch)] = test_loss

        self.after_fold_epoch(epoch, fold, fold_epoch, train_loss, test_loss)
        if verbose > 0:
            print(f"epoch: {epoch}, train loss: {train_loss}, test loss: {test_loss}")

        call_callable_dynamic_args(
            callbacks,
            epoch=epoch, fold=fold, fold_epoch=fold_epoch, loss=train_loss, val_loss=test_loss,
            y_train=train_data.y, y_test=[td.y for td in test_data],
            y_hat_train=LazyInit(lambda: self.predict(train_data.x)),
            y_hat_test=[LazyInit(lambda: self.predict(td.x)) for td in test_data]
        )

    def init_fit(self, **kwargs):
        pass

    def init_fold(self, epoch: int, fold: int):
        pass

    @abstractmethod
    def fit_batch(self, x: pd.DataFrame, y: pd.DataFrame, w: pd.DataFrame, fold: int, **kwargs):
        raise NotImplemented

    def after_fold_epoch(self, epoch, fold, fold_epoch, loss, val_loss):
        pass

    @abstractmethod
    def calculate_loss(self, fold: int, x: pd.DataFrame, y_true: pd.DataFrame, weight: pd.DataFrame) -> float:
        raise NotImplemented

    def merge_folds(self, epoch: int):
        pass

    def train_predict(self, *args, **kwargs) -> Typing.PatchedDataFrame:
        return self.predict(*args, **kwargs)

    @abstractmethod
    def predict(self, features: pd.DataFrame, targets: pd.DataFrame = None, latent: pd.DataFrame = None, samples: int = 1, **kwargs) -> Typing.PatchedDataFrame:
        raise NotImplemented

    def finish_learning(self):
        pass

    def save(self, filename: str):
        """
        save model to disk
        :param filename: filename inclusive file extension
        :return: None
        """
        with open(filename, 'wb') as file:
            pickle.dump(self, file)

        print(f"saved model to: {os.path.abspath(filename)}")

    def plot_loss(self, figsize=(8, 6), **kwargs):
        """
        plot a diagram of the training and validation losses per fold
        :return: figure and axis
        """

        import matplotlib.pyplot as plt
        cm = 'tab20c'  # 'Pastel1'
        df = pd.DataFrame(self._history)
        fig, ax = plt.subplots(1, 1, figsize=(figsize if figsize else plt.rcParams.get('figure.figsize')))
        df['test'].plot(style='--', colormap=cm, ax=ax)
        df['train'].plot(colormap=cm, ax=ax)
        plt.legend(loc='upper right')
        return fig

    def __call__(self, *args, **kwargs):
        """
        returns a copy pf the model with eventually different configuration (kwargs). This is useful for hyper paramter
        tuning or for MultiModels

        :param args:
        :param kwargs: arguments which are eventually provided by hyperopt or by different targets
        :return:
        """
        copy = deepcopy(self)
        copy.kwargs = merge_kwargs(copy.kwargs, kwargs)
        return copy


class AutoEncoderModel(Model):

    # mode constants
    AUTOENCODE = 'autoencode'
    ENCODE = 'encode'
    DECODE = 'decode'

    def __init__(self,
                 features_and_labels: FeaturesAndLabels,
                 summary_provider: Callable[[Typing.PatchedDataFrame], Summary] = Summary,
                 **kwargs):
        super().__init__(features_and_labels, summary_provider, **kwargs)
        self.mode = AutoEncoderModel.AUTOENCODE

    def predict(self, features: pd.DataFrame, targets: pd.DataFrame=None, latent: pd.DataFrame=None, samples=1, **kwargs) -> Typing.PatchedDataFrame:
        if self.mode == AutoEncoderModel.AUTOENCODE:
            return self._auto_encode(features, samples, **kwargs)
        elif self.mode == AutoEncoderModel.ENCODE:
            return self._encode(features, samples, **kwargs)
        elif self.mode == AutoEncoderModel.DECODE:
            return self._decode(latent, samples, **kwargs)
        else:
            raise ValueError("Illegal mode")

    def as_auto_encoder(self) -> 'AutoEncoderModel':
        copy = self()
        copy.mode = AutoEncoderModel.AUTOENCODE
        return copy

    def as_encoder(self) -> 'AutoEncoderModel':
        copy = self()
        copy.mode = AutoEncoderModel.ENCODE
        return copy

    def as_decoder(self) -> 'AutoEncoderModel':
        copy = self()
        copy.mode = AutoEncoderModel.DECODE
        return copy

    @abstractmethod
    def _auto_encode(self, features: pd.DataFrame, samples, **kwargs) -> Typing.PatchedDataFrame:
        raise NotImplemented

    @abstractmethod
    def _encode(self, features: pd.DataFrame, samples, **kwargs) -> Typing.PatchedDataFrame:
        raise NotImplemented

    @abstractmethod
    def _decode(self, latent_features: pd.DataFrame, samples, **kwargs) -> Typing.PatchedDataFrame:
        raise NotImplemented


class SubModelFeature(object):

    def __init__(self, name: str, model: Model):
        self.name = name
        self.model = model

    def fit(self, df: Typing.PatchedDataFrame, **kwargs):
        _log.info(f"fitting submodel: {self.name}")
        with df.model() as m:
            fit = m.fit(self.model, **kwargs)
            self.model = fit.model

        _log.info(f"fitted submodel: {fit}")
        return self.predict(df, **kwargs)

    def predict(self, df: Typing.PatchedDataFrame, **kwargs):
        if isinstance(self.model, AutoEncoderModel):
            return df.model.predict(self.model.as_encoder(), **kwargs)[PREDICTION_COLUMN_NAME]
        else:
            return df.model.predict(self.model, **kwargs)[PREDICTION_COLUMN_NAME]




from unittest import TestCase

import pandas as pd
import numpy as np

from pandas_ml_utils.ml.summary.figures import df_regression_scores, df_classification_scores
from pandas_ml_utils_test.config import DF_SUMMARY_REGRESSION, DF_SUMMARY_CLASSIFICATION


class TestFigures(TestCase):

    def test_regression_figures(self):
        print(df_regression_scores(DF_SUMMARY_REGRESSION, None, no_style=True))
        self.assertFalse(df_regression_scores(DF_SUMMARY_REGRESSION, None, no_style=True).isnull().values.any())

    def test_classification_figures(self):
        print(df_classification_scores(DF_SUMMARY_CLASSIFICATION, None, no_style=True))
        #print(df_classification_scores(DF_SUMMARY_CLASSIFICATION, None)._repr_html_())
        self.assertNotEqual(df_classification_scores(DF_SUMMARY_CLASSIFICATION, None, no_style=True).max().item(), np.nan)




from unittest import TestCase

from sklearn.neural_network import MLPRegressor

from pandas_ml_utils import SkModel, FeaturesAndLabels, FittingParameter
from pandas_ml_utils.ml.callback import EarlyStopping
from pandas_ml_utils_test.config import DF_AB_10


class TestCallBack(TestCase):

    def test_early_stopping(self):
        """given early stopping callback"""
        cb = EarlyStopping(2, tolerance=-10)

        """when training is not improveing"""
        with DF_AB_10.model() as m:
            m.fit(
                SkModel(
                    MLPRegressor([1, 2, 1]),
                    FeaturesAndLabels(["a"], ["b"]),
                ),
                FittingParameter(epochs=50),
                verbose=1,
                callbacks=cb
            )

        """then not all epochs were executed"""
        self.assertLess(cb.call_counter, 50)
        self.assertGreaterEqual(cb.call_counter, cb.patience)


from collections import namedtuple
from typing import List

from pandas_ml_common import Typing
from pandas_ml_common.utils.serialization_utils import plot_to_html_img
from pandas_ml_utils import html
from .figures import *


class Summary(object):
    """
    Summary objects a used to visually present the results of a `df.fit` fitted model or a `df.backtest`
    All implementations of `Summary` need to:
     * pass a `pd.DataFrame` to `super().__init()__`
     * implement `_repr_html_()`
    """

    Cell = namedtuple("Cell", ["index", "rowspan", "colspan"])

    def __init__(self, df: Typing.PatchedDataFrame, model: 'Model', *args, layout: List[List[int]] = None, **kwargs):
        self._df = df
        self.model = model
        self.args = args
        self.kwargs = kwargs

        if layout is not None:
            grid = np.array(layout, dtype=int)
            table: List[List[Summary.Cell]] = [[]]
            for i, j in np.ndindex(grid.shape):
                if i > 0:
                    if grid[i-1, j] == grid[i, j]:
                        # rowspan += 1
                        cell = table[-1][j]
                        table[-1][j] = Summary.Cell(cell.index, cell.rowspan + 1, cell.colspan)

                    if j <= 0:
                        # we need to add a new empty row
                        table.append([])
                if j > 0:
                    if grid[i, j-1] == grid[i, j]:
                        # colspan += 1
                        cell = table[i][-1]
                        table[i][-1] = Summary.Cell(cell.index, cell.rowspan, cell.colspan + 1)
                    else:
                        table[i].append(Summary.Cell(grid[i, j], 1, 1))
                else:
                    table[i].append(Summary.Cell(grid[i, j], 1, 1))

            # assign layout
            if len(table) > 0 and table[-1] == []:
                table.pop()

            self.layout = table
            self.laypout_nr_columns = grid.shape[1]
        else:
            self.layout = None
            self.laypout_nr_columns = None

    @property
    def df(self):
        return self._df

    def __str__(self):
        return str(self.df.groupby(level=0).tail(1)) if isinstance(self.df.index, pd.MultiIndex) else str(self.df.tail())

    def _repr_html_(self):
        from mako.template import Template
        from mako.lookup import TemplateLookup
        plot = "<class 'matplotlib.figure.Figure'>"

        figures = [arg(self.df, model=self.model) for arg in self.args]
        figures = [("img", plot_to_html_img(f)) if str(type(f)) == plot else ("html", f._repr_html_() if hasattr(f, "_repr_html_") else str(f)) for f in figures]

        template = Template(filename=html.SELF_TEMPLATE(__file__), lookup=TemplateLookup(directories=['/']))
        return template.render(summary=self, figures=figures, layout=self.layout, laypout_nr_columns=self.laypout_nr_columns)


class RegressionSummary(Summary):

    def __init__(self, df: Typing.PatchedDataFrame, model: 'Model', **kwargs):
        super().__init__(
            df,
            model,
            plot_true_pred_scatter,
            df_regression_scores,
            plot_feature_importance,
            df_tail,
            layout=[[0, 1],
                    [2, 2],
                    [3, 3]],
            **kwargs
        )

    def __str__(self):
        return f"{self.args[1](self.df, model=self.model)}\n{super().__str__()}"


class ClassificationSummary(Summary):

    def __init__(self, df: Typing.PatchedDataFrame, model: 'Model', include_feature_importance=True, **kwargs):
        super().__init__(
            df,
            model,
            plot_confusion_matrix,                          # 0
            plot_receiver_operating_characteristic,         # 1
            df_classification_scores,                       # 2
            plot_feature_importance,                        # 3
            df_tail,                                        # 4
            layout=([[2, 0, 0, 1, 1],
                     [3, 3, 3, 3, 3],
                     [4, 4, 4, 4, 4]] if include_feature_importance else [[2, 0, 0, 1, 1],
                                                                          [4, 4, 4, 4, 4]]),
            **kwargs
        )

    def __str__(self):
        from mlxtend.evaluate import confusion_matrix

        # get true and prediction data. It needs to be a one hot encoded 2D array [samples, class] where nr_classes >= 2
        tv, pv = clean_one_hot_classification(self.df[LABEL_COLUMN_NAME]._.values, self.df[PREDICTION_COLUMN_NAME]._.values)

        # confusion matrix needs integer encoding
        tv = np.apply_along_axis(np.argmax, 1, tv)
        pv = np.apply_along_axis(np.argmax, 1, pv)
        cm = confusion_matrix(tv, pv, binary=tv.max() < 2)

        return f"{cm}"




from __future__ import annotations

import logging
from copy import deepcopy
from typing import Callable, Tuple, List, Union, Generator

import numpy as np
import pandas as pd
from sklearn.exceptions import ConvergenceWarning
from sklearn.linear_model import LogisticRegression
from sklearn.neural_network import MLPRegressor
from sklearn.base import clone, ClusterMixin, ClassifierMixin, RegressorMixin
from sklearn import metrics

from pandas_ml_common import Typing
from pandas_ml_common.utils import call_callable_dynamic_args, unpack_nested_arrays, merge_kwargs, to_pandas
from pandas_ml_common.utils.logging_utils import LogOnce
from pandas_ml_utils.ml.data.extraction import FeaturesAndLabels
from pandas_ml_utils.ml.summary import Summary
from .base_model import Model, AutoEncoderModel

_log = logging.getLogger(__name__)
ConvergenceWarning('ignore')


class _AbstractSkModel(Model):

    def __init__(self,
                 skit_model,
                 features_and_labels: FeaturesAndLabels,
                 summary_provider: Callable[[Typing.PatchedDataFrame], Summary] = Summary,
                 **kwargs):
        super().__init__(features_and_labels, summary_provider, **kwargs)
        self.sk_model = skit_model
        self.overwrite_folds = True  # maybe later we can handle the folding of models (eventually just taking the best)
        self.log_once = LogOnce().log
        self._label_shape = None

    def init_fold(self, epoch: int, fold: int):
        if fold > 1:
            self.log_once("merge_folds", _log.warning,
                          "merging of cross folded models is not supported, we just keep training the same model")

    def fit_batch(self, x: pd.DataFrame, y: pd.DataFrame, weight: pd.DataFrame, fold: int, **kwargs):
        # convert data frames to numpy arrays
        _x = _AbstractSkModel.reshape_rnn_as_ar(unpack_nested_arrays(x, split_multi_index_rows=False))
        _y = unpack_nested_arrays(y, split_multi_index_rows=False)
        _w = unpack_nested_arrays(weight, split_multi_index_rows=False)

        _y = _y.reshape((len(_x), -1)) if _y.ndim > 1 and _y.shape[1] == 1 else _y
        _y = _y.reshape(len(_x)) if _y.ndim == 2 and _y.shape[1] == 1 else _y
        if self._label_shape is None: self._label_shape = _y.shape

        par = self._fit_meta_data
        partial_fit = any([size > 1 for size in [par.epochs, par.batch_size, par.fold_epochs] if size is not None])

        if partial_fit:
            # use partial fit whenever possible partial_fit
            if hasattr(self.sk_model, "partial_fit"):
                kw_classes = {"classes": kwargs["classes"]} if "classes" in kwargs else {}
                try:
                    self.sk_model = self.sk_model.partial_fit(_x, _y, **kw_classes)
                except Exception as e:
                    if "classes" in kwargs:
                        raise e
                    else:
                        raise ValueError("You might need to pass 'classes' argument for partial fitting", e)
            else:
                raise ValueError(f"This of model does not support `partial_fit` {type(self.sk_model)} - "
                                 f"and therefore does not support epochs or batches.")
        else:
            self.sk_model = self.sk_model.fit(_x, _y)

    def merge_folds(self, epoch: int):
        pass

    def finish_learning(self):
        # clear oll intermediate fold models, otherwise they get serialized to disk
        pass

    @staticmethod
    def reshape_rnn_as_ar(arr3d):
        if arr3d.ndim < 3:
            return arr3d
        else:
            return arr3d.reshape(arr3d.shape[0], np.array(arr3d.shape[1:]).prod())


class SkModel(_AbstractSkModel):

    def __init__(self,
                 skit_model,
                 features_and_labels: FeaturesAndLabels,
                 summary_provider: Callable[[Typing.PatchedDataFrame], Summary] = Summary,
                 **kwargs):
        super().__init__(skit_model, features_and_labels, summary_provider, **kwargs)

    def calculate_loss(self, fold, x, y_true, weight):
        skm = self.sk_model
        y_pred = self._predict(skm, x, fold=fold)
        y_true = unpack_nested_arrays(y_true, split_multi_index_rows=False).reshape(y_pred.shape)
        w = weight.values.reshape(-1, ) if weight is not None else None

        if isinstance(self.sk_model, ClassifierMixin):
            # calculate: # sklearn.metrics.log_loss
            return metrics.log_loss(y_true, y_pred, sample_weight=w)
        else:
            # calculate: metrics.mean_squared_error
            return metrics.mean_squared_error(y_true, y_pred, sample_weight=w)

    def predict(self, features: pd.DataFrame, targets: pd.DataFrame=None, latent: pd.DataFrame=None, samples=1, **kwargs) -> Typing.PatchedDataFrame:
        return to_pandas(self._predict(self.sk_model, features, samples, **kwargs), features.index, self._labels_columns)

    def _predict(self, skm, features: pd.DataFrame, samples=1, **kwargs) -> np.ndarray:
        x = _AbstractSkModel.reshape_rnn_as_ar(unpack_nested_arrays(features, split_multi_index_rows=False))
        is_probabilistic = callable(getattr(skm, 'predict_proba', None))

        def predictor():
            if is_probabilistic:
                y_hat = skm.predict_proba(x)
                binary_classifier = len(self._label_shape) == 1 or self._label_shape[1] == 1
                return (1 - y_hat[:, 0]) if binary_classifier else y_hat.reshape(-1, *self._label_shape[1:])
            else:
                return skm.predict(x)

        return np.array([predictor() for _ in range(samples)]).swapaxes(0, 1) if samples > 1 else predictor()


class SkAutoEncoderModel(_AbstractSkModel, AutoEncoderModel):

    def __init__(self,
                 encode_layers: List[int],
                 decode_layers: List[int],
                 features_and_labels: FeaturesAndLabels,
                 summary_provider: Callable[[Typing.PatchedDataFrame], Summary] = Summary,
                 **kwargs):
        super().__init__(
            skit_model=call_callable_dynamic_args(MLPRegressor, **{"hidden_layer_sizes": [*encode_layers, *decode_layers], **kwargs}),
            features_and_labels=features_and_labels,
            summary_provider=summary_provider,
            **kwargs
        )

        # Implementation analog blog: https://i-systems.github.io/teaching/ML/iNotes/15_Autoencoder.html
        self.encoder_layers = encode_layers
        self.decoder_layers = decode_layers
        self.layers = [*encode_layers, *decode_layers]

    def calculate_loss(self, fold, x, y_true, weight) -> float:
        skm = self.sk_model
        y_pred = skm.predict(_AbstractSkModel.reshape_rnn_as_ar(unpack_nested_arrays(x)))
        y_true = unpack_nested_arrays(y_true, split_multi_index_rows=False).reshape(y_pred.shape)
        w = weight.values.reshape(-1, ) if weight is not None else None

        return metrics.mean_squared_error(y_true, y_pred, sample_weight=w)


    def _auto_encode(self, features: pd.DataFrame, samples, **kwargs) -> Typing.PatchedDataFrame:
        x = _AbstractSkModel.reshape_rnn_as_ar(unpack_nested_arrays(features, split_multi_index_rows=False))
        return to_pandas(self.sk_model.predict(x), features.index, self._labels_columns)

    def _encode(self, features: pd.DataFrame, samples, **kwargs) -> Typing.PatchedDataFrame:
        skm = self.sk_model
        if not hasattr(skm, 'coefs_'):
            raise ValueError("Model needs to be 'fit' first!")

        encoder = call_callable_dynamic_args(MLPRegressor, **{"hidden_layer_sizes": self.encoder_layers[1:], **self.kwargs})
        encoder.coefs_ = skm.coefs_[:len(self.encoder_layers)].copy()
        encoder.intercepts_ = skm.intercepts_[:len(self.encoder_layers)].copy()
        encoder.n_layers_ = len(encoder.coefs_) + 1
        encoder.n_outputs_ = len(self.features_and_labels.latent_names)
        encoder.out_activation_ = skm.activation

        encoded = encoder.predict(_AbstractSkModel.reshape_rnn_as_ar(unpack_nested_arrays(features, split_multi_index_rows=False)))
        return to_pandas(encoded, features.index, self._features_and_labels.latent_names)

    def _decode(self, latent_features: pd.DataFrame, samples, **kwargs) -> Typing.PatchedDataFrame:
        skm = self.sk_model
        if not hasattr(skm, 'coefs_'):
            raise ValueError("Model needs to be 'fit' first!")

        decoder = call_callable_dynamic_args(MLPRegressor, **{"hidden_layer_sizes": self.decoder_layers, **self.kwargs})
        decoder.coefs_ = skm.coefs_[len(self.encoder_layers):].copy()
        decoder.intercepts_ = skm.intercepts_[len(self.encoder_layers):].copy()
        decoder.n_layers_ = len(decoder.coefs_) + 1
        decoder.n_outputs_ = self.layers[-1]
        decoder.out_activation_ = skm.out_activation_

        decoded = decoder.predict(_AbstractSkModel.reshape_rnn_as_ar(unpack_nested_arrays(latent_features, split_multi_index_rows=False)))
        return to_pandas(decoded, latent_features.index, self._feature_columns)




from typing import Union, Tuple

import numpy as np
import scipy.stats as st
from pandas_ml_common import Typing

class NormalConfidence(object):

    def __init__(self, confidence: Union[float, Tuple[float, float]] = 0.95,):
        self.confidence = (
            np.abs(st.norm.ppf(confidence[0])),  # i.e 0.025
            np.abs(st.norm.ppf(confidence[1]))   # i.e 0.972
        ) if isinstance(confidence, tuple) else (
            np.abs(st.norm.ppf((1. - confidence) / 2.)),
        ) * 2

    def band(self, value, std=None):
        return self.lower(value, std), self.upper(value, std)

    def lower(self, value, std=None):
        value, std = self._extract_args(value, std)
        l = (value - std * self.confidence[0])
        return l.values.squeeze() if isinstance(l, Typing.AnyPandasObject) else l

    def upper(self, value, std=None):
        value, std = self._extract_args(value, std)
        u = (value + std * self.confidence[1])
        return u.values.squeeze() if isinstance(u, Typing.AnyPandasObject) else u

    def _extract_args(self, value, std):
        if std is None:
            if value.ndim >= 1:
                if isinstance(value, Typing.AnyPandasObject):
                    value = value._.values.squeeze()

                return value[0], value[1]
            else:
                raise ValueError("you need to pass a value, std tuple or value, std as arguments")
        else:
            return value, std

    def __call__(self, value, std=None):
        return self.band(value, std)


from pandas_ml_utils import FittingParameter
from pandas_ml_utils_test.ml.model.test_model_accuracy import TestModelAccuracy


class TestSkModelAccuracy(TestModelAccuracy):

    def provide_linear_regression_model(self):
        from sklearn.linear_model import LinearRegression
        from sklearn.neural_network import MLPRegressor
        from pandas_ml_utils import FeaturesAndLabels, SkModel

        return [
            (
                SkModel(LinearRegression(), FeaturesAndLabels(["x"], ["y"])),
                FittingParameter(epochs=1, fold_epochs=1, context="LinearRegression")
            ),
            (
                SkModel(
                    MLPRegressor(10, learning_rate_init=0.01, max_iter=9000, validation_fraction=0),
                    FeaturesAndLabels(["x"], ["y"])
                 ),
                FittingParameter(epochs=1, fold_epochs=1, context="MLPRegressor")
            ),
            (
                SkModel(
                    MLPRegressor(10, learning_rate_init=0.01, max_iter=1, validation_fraction=0, warm_start=True),
                    FeaturesAndLabels(["x"], ["y"])
                 ),
                FittingParameter(epochs=9000, fold_epochs=1, context="MLPRegressor partial fit")
            )
        ]

    def provide_non_linear_regression_model(self):
        from sklearn.neural_network import MLPRegressor
        from pandas_ml_utils import FeaturesAndLabels, SkModel

        return [
            (
                SkModel(
                    MLPRegressor(200, learning_rate_init=0.001, max_iter=5000, validation_fraction=0),
                    FeaturesAndLabels(["x"], ["y"])
                ),
                FittingParameter(epochs=1, context="epoch 1 fit"),
            ),
            (
                SkModel(
                    MLPRegressor(200, learning_rate_init=0.001, max_iter=1, validation_fraction=0, warm_start=True),
                    FeaturesAndLabels(["x"], ["y"])
                ),
                FittingParameter(epochs=5000, context="partial fit"),
            )
        ]






import os
import tempfile
import uuid
from typing import Tuple

from pandas_ml_common import pd, np, naive_splitter, random_splitter
from pandas_ml_utils import FeaturesAndLabels, Model, SubModelFeature, FittingParameter
from pandas_ml_utils.ml.model.base_model import AutoEncoderModel


class TestAbstractModel(object):

    def test_classifier(self):
        """given some toy classification data"""
        df = pd.DataFrame({
            "a": [1, 0, 1, 0, 1, 0, 1, 0,],
            "b": [0, 0, 1, 1, 0, 0, 1, 1,],
            "c": [1, 0, 0, 1, 1, 0, 0, 1,]
        })

        """and a model"""
        model = self.provide_classification_model(FeaturesAndLabels(features=["a", "b"], labels=["c"], label_type=int))
        temp = os.path.join(tempfile.gettempdir(), str(uuid.uuid4()))

        """when we fit the model"""
        batch_size, epochs = self.provide_batch_size_and_epoch()
        with df.model(temp) as m:
            fit = m.fit(model, FittingParameter(splitter=naive_splitter(0.49), batch_size=batch_size, epochs=epochs), verbose=0)

        print(fit.training_summary.df)
        # fit.training_summary.df.to_pickle('/tmp/classifier.df')
        # print(fit._repr_html_())

        """then we get a html summary and can predict"""
        self.assertIn('<style>', fit.training_summary._repr_html_())

        prediction = df.model.predict(fit.model)
        binary_prediction = prediction.iloc[:,0] >= 0.5
        np.testing.assert_array_equal(binary_prediction, np.array([True, False, False, True, True, False, False, True,]))

        """and load the model"""
        try:
            copy = Model.load(temp)
            pd.testing.assert_frame_equal(df.model.predict(fit.model), df.model.predict(copy), check_less_precise=True)
        finally:
            os.remove(temp)

    def test_regressor(self):
        """given some toy regression data"""
        df = pd.DataFrame({
            "a": [-1.0, 0.0, 1.0, 2.0, 3.0, 4.0],
            "b": [-2.0, 1.0, 4.0, 7.0, 10.0, 13.0]
        })

        """and a model"""
        model = self.provide_regression_model(FeaturesAndLabels(features=["a"], labels=["b"]))

        """when we fit the model"""
        batch_size, epochs = self.provide_batch_size_and_epoch()
        with df.model() as m:
            fit = m.fit(model, FittingParameter(splitter=naive_splitter(0.3), batch_size=batch_size, epochs=epochs),
                        verbose=0)

        print(fit.training_summary.df)
        self.assertEqual(4, len(fit.training_summary.df))
        self.assertEqual(2, len(fit.test_summary.df))

        """then we can predict"""
        prediction = df.model.predict(fit.model)
        np.testing.assert_array_almost_equal(prediction.iloc[:, 0].values, df["b"].values, 1)

        """and save and load the model"""
        temp = os.path.join(tempfile.gettempdir(), str(uuid.uuid4()))
        try:
            fit.model.save(temp)
            copy = Model.load(temp)
            pd.testing.assert_frame_equal(df.model.predict(fit.model), df.model.predict(copy), check_less_precise=True)
        finally:
            os.remove(temp)

    def test_auto_encoder(self):
        """given the implementation can handle auto encoders"""
        model = self.provide_auto_encoder_model(
            FeaturesAndLabels(
                features=["a", "b"],
                labels=["a", "b"],
                latent=["x"]
            )
        )

        if model is None:
            return

        """and some toy classification data"""
        df = pd.DataFrame({
            "a": [1, 0] * 10,
            "b": [0, 1] * 10,
        })

        """when we fit the model"""
        batch_size, epochs = self.provide_batch_size_and_epoch()
        with df.model() as m:
            fit = m.fit(model, FittingParameter(splitter=naive_splitter(0.49), batch_size=batch_size, epochs=epochs),
                        verbose=0)

        print(fit.training_summary.df)

        """then we can predict Autoencoded"""
        auto_encoded_prediction = df.model.predict(fit.model)
        self.assertEqual((20, 2), auto_encoded_prediction["prediction"].shape)

        """and we can encode"""
        encoded_prediction = df.model.predict(fit.model.as_encoder())
        print(encoded_prediction)
        self.assertEqual((20, 1), encoded_prediction["prediction"].shape)

        """and we can decode"""
        decoded_prediction = encoded_prediction["prediction"].model.predict(fit.model.as_decoder())
        print(decoded_prediction)
        np.testing.assert_array_almost_equal(decoded_prediction["prediction"].values > 0.5, df[["a", "b"]].values)

        """and we can encoder and decode after safe and load"""
        temp = os.path.join(tempfile.gettempdir(), str(uuid.uuid4()))
        try:
            fit.model.save(temp)
            copy = Model.load(temp)

            pd.testing.assert_frame_equal(
                df.model.predict(fit.model.as_encoder()),
                df.model.predict(copy.as_encoder()),
                check_less_precise=True)

            pd.testing.assert_frame_equal(
                encoded_prediction.model.predict(fit.model.as_decoder()),
                encoded_prediction.model.predict(copy.as_decoder()),
                check_less_precise=True)
        finally:
            os.remove(temp)

        # try to save only as encoder model
        try:
            fit.model.as_encoder().save(temp)
            copy = Model.load(temp)
        finally:
            os.remove(temp)

    def test_multi_sample_regressor(self):
        """given some toy regression data"""
        df = pd.DataFrame({
            "a": [-1.0, 0.0, 1.0, 2.0, 3.0, 4.0],
            "b": [-2.0, 1.0, 4.0, 7.0, 10.0, 13.0]
        })

        """and a model"""
        model = self.provide_regression_model(FeaturesAndLabels(features=["a"], labels=["b"]))

        """when we fit the model"""
        batch_size, epochs = self.provide_batch_size_and_epoch()
        with df.model() as m:
            fit = m.fit(model, FittingParameter(splitter=naive_splitter(0.3), batch_size=batch_size, epochs=epochs),
                        verbose=0)

        print(fit.training_summary.df)

        """then we can predict"""
        prediction = df.model.predict(fit.model, samples=2)
        np.testing.assert_array_almost_equal(prediction.iloc[:, 0]._.values, np.concatenate([df[["b"]].values, df[["b"]].values], axis=1), 1)

    def test_multindex_row(self):
        """given some toy regression data while we provide a multiindex for the rows"""
        df = pd.DataFrame({
            "a": [-1.0, 0.0, 1.0, 2.0, 3.0, 4.0, -1.0, 0.0, 1.0, 2.0, 3.0, 4.0],
            "b": [-2.0, 1.0, 4.0, 7.0, 10.0, 13.0, -2.0, 1.0, 4.0, 7.0, 10.0, 13.0]
        }, index=pd.MultiIndex.from_product([["A", "B"], range(6)]))

        """and a model"""
        model = self.provide_regression_model(FeaturesAndLabels(features=["a"], labels=["b"]))

        """when we fit the model"""
        batch_size, epochs = self.provide_batch_size_and_epoch()
        with df.model() as m:
            fit = m.fit(model,
                        FittingParameter(splitter=random_splitter(0.3, partition_row_multi_index=True), batch_size=batch_size, epochs=epochs),
                        verbose=0)

        prediction = df.model.predict(fit.model)
        print(fit)
        # fit.training_summary.df.to_pickle('/tmp/multi_index_row_summary.df')
        # print(fit._repr_html_())

        """then we get a prediction for A and B rows"""
        self.assertEqual(8, len(fit.training_summary.df))
        self.assertEqual(4, len(fit.training_summary.df.loc["A"]))
        self.assertEqual(4, len(fit.training_summary.df.loc["B"]))

        self.assertEqual(4, len(fit.test_summary.df))
        self.assertEqual(2, len(fit.test_summary.df.loc["A"]))
        self.assertEqual(2, len(fit.test_summary.df.loc["B"]))

        self.assertEqual(6, len(prediction.loc["A"]))
        self.assertEqual(6, len(prediction.loc["B"]))
        np.testing.assert_array_almost_equal(prediction.iloc[:, 0].values, df["b"].values, 1)

    def test_multindex_row_multi_samples(self):
        """given some toy regression data while we provide a multiindex for the rows"""
        df = pd.DataFrame({
            "a": [-1.0, 0.0, 1.0, 2.0, 3.0, 4.0, -1.0, 0.0, 1.0, 2.0, 3.0, 4.0],
            "b": [-2.0, 1.0, 4.0, 7.0, 10.0, 13.0, -2.0, 1.0, 4.0, 7.0, 10.0, 13.0]
        }, index=pd.MultiIndex.from_product([["A", "B"], range(6)]))

        """and a model"""
        model = self.provide_regression_model(FeaturesAndLabels(features=["a"], labels=["b"]))

        """when we fit the model"""
        batch_size, epochs = self.provide_batch_size_and_epoch()
        with df.model() as m:
            fit = m.fit(model,
                        FittingParameter(splitter=random_splitter(0.3, partition_row_multi_index=True), batch_size=batch_size, epochs=epochs),
                        verbose=0)

        self.assertEqual(8, len(fit.training_summary.df))
        self.assertEqual(4, len(fit.test_summary.df))

        prediction = df.model.predict(fit.model, samples=2)
        self.assertEqual(2, len(prediction.iloc[:, 0]._.values))
        self.assertEqual((6, 2), prediction.loc["A"].iloc[:, 0]._.values.shape)
        self.assertEqual((6, 2), prediction.loc["B"].iloc[:, 0]._.values.shape)

    def test_no_test_data(self):
        """given some toy regression data"""
        df = pd.DataFrame({
            "a": [-1.0, 0.0, 1.0, 2.0, 3.0, 4.0],
            "b": [-2.0, 1.0, 4.0, 7.0, 10.0, 13.0]
        })

        """and a model"""
        model = self.provide_regression_model(FeaturesAndLabels(features=["a"], labels=["b"]))

        """when we fit the model"""
        batch_size, epochs = self.provide_batch_size_and_epoch()
        with df.model() as m:
            fit = m.fit(model,
                        FittingParameter(splitter=naive_splitter(0), batch_size=batch_size, epochs=epochs),
                        verbose=0)

        # print(fit.training_summary.df)
        print(fit.test_summary.df)

        """then we have an empty test data frame"""
        self.assertEqual(len(fit.training_summary.df), len(df))
        self.assertEqual(len(fit.test_summary.df), 0)

    def test_stacked_models(self):
        """given some toy classification data"""
        df = pd.DataFrame({
            "a": [1, 0, 1, 0, 1, 0, 1, 0, ],
            "b": [0, 0, 1, 1, 0, 0, 1, 1, ],
            "c": [1, 0, 0, 1, 1, 0, 0, 1, ]
        })

        """and a model"""
        model = self.provide_classification_model(
            FeaturesAndLabels(
                features=[
                    "a",
                    SubModelFeature("b", self.provide_classification_model(
                        FeaturesAndLabels(features=["a", "b"], labels=["c"], label_type=int)
                    ))
                ],
                labels=["c"],
                label_type=int
            )
        )

        temp = os.path.join(tempfile.gettempdir(), str(uuid.uuid4()))

        with self.assertLogs(level='INFO') as cm:
            with df.model(temp) as m:
                fit = m.fit(model)

            self.assertIn("INFO:pandas_ml_utils.ml.model.base_model:fitting submodel: b", cm.output[0])
            self.assertIn(
                "INFO:pandas_ml_utils.ml.model.base_model:fitted submodel",
                [s for s in cm.output if s.startswith("INFO:pandas_ml_utils.ml.model.base_model:fitted")][0]
            )

        prediction = df.model.predict(fit.model)
        prediction2 = df.model.predict(Model.load(temp))
        pd.testing.assert_frame_equal(prediction, prediction2)
        os.remove(temp)

    # Abstract methods
    def provide_batch_size_and_epoch(self) -> Tuple[int, int]:
        return (None, 1)

    def provide_regression_model(self, features_and_labels) -> Model:
        pass

    def provide_classification_model(self, features_and_labels) -> Model:
        pass

    def provide_auto_encoder_model(self, features_and_labels) -> AutoEncoderModel:
        return None


import os

import numpy as np

from pandas_ml_common import pd

TEST_DATA_PATH = os.path.join(os.path.dirname(os.path.abspath(__file__)), ".data")
DF_NOTES = pd.read_csv(os.path.join(TEST_DATA_PATH, "banknote_authentication.csv"))
DF_DEBUG = pd.DataFrame({"Close": np.random.random(10)})
DF_AB_10 = pd.DataFrame({"a": np.random.random(10), "b": np.random.random(10)})
DF_TEST = pd.read_csv(os.path.join(TEST_DATA_PATH, "SPY.csv"), index_col='Date')
DF_SUMMARY_REGRESSION = pd.read_pickle(os.path.join(TEST_DATA_PATH, "multi_index_row_summary.df"))  # generated by: test_multindex_row
DF_SUMMARY_CLASSIFICATION = pd.read_pickle(os.path.join(TEST_DATA_PATH, "classifier.df"))  # test_classifier


from datetime import datetime
from functools import wraps
from string import Template
from typing import Union

from pandas_ml_common import Typing
from pandas_ml_common.utils.time_utils import seconds_since_midnight
from pandas_ml_utils.df_patching.model_patch import DfModelPatch
from pandas_ml_utils.ml.data.extraction.features_and_labels_definition import FeaturesAndLabels
from pandas_ml_utils.ml.data.extraction.features_and_labels_extractor import FeaturesWithLabels
from pandas_ml_utils.ml.fitting import Fit
from pandas_ml_utils.ml.model import Model


class ModelContext(object):

    def __init__(self, df: Typing.PatchedDataFrame, file_name: str = None):
        self.df = df
        self.file_name = file_name

    @wraps(DfModelPatch.fit)
    def fit(self, *args, **kwargs) -> Fit:
        fit = self.df.model.fit(*args, **kwargs)

        if self.file_name is not None:
            file_name = Template(self.file_name).substitute({
                **kwargs,
                "V": f'{datetime.now().strftime("%Y%m-%d")}-{seconds_since_midnight()}'
            })

            fit.model.save(file_name)

        return fit

    def extract(self, model_or_fnl: Union[Model, FeaturesAndLabels], **kwargs) -> FeaturesWithLabels:
        if isinstance(model_or_fnl, Model):
            kwargs = {**model_or_fnl.features_and_labels.kwargs, **model_or_fnl.kwargs, **kwargs}
            return self.df._.extract(model_or_fnl.features_and_labels, **kwargs)
        else:
            kwargs = {**model_or_fnl.kwargs, **kwargs}
            return self.df._.extract(model_or_fnl, **kwargs)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        if exc_type:
            print(f'exc_type: {exc_type}')
            print(f'exc_value: {exc_value}')
            print(f'exc_traceback: {exc_traceback}')

            # eventually return True if all excetions are handled




from collections import namedtuple
from typing import List

import numpy as np
import pandas as pd
from sklearn.ensemble._forest import ForestClassifier, ForestRegressor
from sklearn.feature_selection import RFECV
from sklearn.model_selection import RandomizedSearchCV

from pandas_ml_common import Typing, Sampler, XYWeight
from pandas_ml_common.utils.numpy_utils import clean_one_hot_classification
from pandas_ml_common.utils.serialization_utils import plot_to_html_img
from pandas_ml_utils import html
from pandas_ml_utils.constants import *
from .figures import *
from .base_summary import Summary

_CONTEXT = "This summary should only be used by df.model.feature_selection()"


class FeatureSelectionSummary(Summary):

    def __init__(self, df: Typing.PatchedDataFrame, model: 'pandas_ml_utils.Model', is_test: bool, **kwargs):
        # assert the summary is used properly i.e. from the feature selection call
        assert hasattr(model, 'sk_model'), _CONTEXT
        assert isinstance(model.sk_model, RandomizedSearchCV), _CONTEXT
        assert isinstance(model.sk_model.best_estimator_, RFECV), _CONTEXT
        assert isinstance(model.sk_model.best_estimator_.estimator_, (ForestClassifier, ForestRegressor)), _CONTEXT

        # init super
        super().__init__(
            df,
            model,
            self._plot_feature_importance,
            plot_feature_correlation,
            lambda *args, **kwargs: self.kpis,
            layout=[
                [0, 0],
                [1, 2],
            ],
            **kwargs
        )

        # init self
        self.is_test = is_test
        self.nr_of_nodes = sum([e.tree_.node_count for e in model.sk_model.best_estimator_.estimator_.estimators_])

        def get_kpis(model):
            # originally passed features
            features = df[FEATURE_COLUMN_NAME].columns

            # get rankings
            rankings = model.sk_model.best_estimator_.ranking_  # best is least aka 1
            df_kpi = pd.DataFrame({"ranking": rankings, "importance": None, "importance_std": None}, index=features)
            selected_features = df_kpi[df_kpi["ranking"] <= 1].index

            # get importance of selected features
            importances = model.sk_model.best_estimator_.estimator_.feature_importances_
            importance_indices = np.argsort(importances)[::-1]

            # sorted kpi's by ranking
            importances = importances[importance_indices]
            importance_names = selected_features[importance_indices]
            importance_std = np.std(
                [tree.feature_importances_ for tree in model.sk_model.best_estimator_.estimator_.estimators_], axis=0)[
                importance_indices]

            # update the feature importance in then data frame
            df_kpi.loc[importance_names, "importance"] = importances
            df_kpi.loc[importance_names, "importance_std"] = importance_std

            return df_kpi

        # assign training data KPIs
        self.kpis = get_kpis(model)

        if is_test:
            from pandas_ml_utils.ml.fitting import FittingParameter

            # if this is the summary of the test set we should re-fit the model and finally compare the feature ranks
            features, labels = df[FEATURE_COLUMN_NAME], df[LABEL_COLUMN_NAME]
            weights = df[SAMPLE_WEIGHTS_COLUMN_NAME] if SAMPLE_WEIGHTS_COLUMN_NAME in df else None

            test_model = model()
            test_model.fit(XYWeight(features, labels, weights), FittingParameter(splitter=None, epochs=1))
            validation_kpis = get_kpis(test_model)

            self.kpis = self.kpis.join(validation_kpis, how='outer', rsuffix=" (Test)")

    def __str__(self):
        return str(self.kpis)

    def _plot_feature_importance(self, *args, figsize=(20, 10), **kwargs):
        fidf = self.kpis.fillna(0)

        if self.is_test:
            ax = fidf[["importance", "importance (Test)"]].plot.bar(
                yerr=fidf[["importance_std", "importance_std (Test)"]].values.T,
                figsize=figsize
            )
        else:
            ax = fidf[["importance", "importance_std"]].plot.bar(yerr='importance_std', figsize=figsize)

        ax.set_xticklabels(ax.get_xticklabels(), ha='right', rotation=60, rotation_mode="anchor")
        return ax.get_figure()

        # PS the single version is:
        # fidf[["importances", "importance_std"]].plot.bar(yerr="importance_std")



from functools import partial
from unittest import TestCase

import numpy as np
import pandas as pd

from pandas_ml_utils.ml.confidence import NormalConfidence


class TestNormalConfidence(TestCase):

    def test_df_confidence(self):
        std = 0.3
        conf = 0.75
        df = pd.DataFrame({"a": np.random.normal(0, std, 1500)})
        x = pd.Series(np.zeros(len(df)), index=df.index).to_frame().apply(partial(NormalConfidence(conf), std=std), result_type='expand', axis=1)
        tail_events = ((df["a"] >= x[0]) & (df["a"] <= x[1])).values.sum()
        tail_events = tail_events / len(df)
        self.assertGreaterEqual(tail_events, conf - 0.02)



from unittest import TestCase

import pandas as pd

from pandas_ml_utils.constants import *
from pandas_ml_utils.ml.data.reconstruction import map_prediction_to_target


class TestReconstruction(TestCase):

    def test_map_prediction_to_target_1o1(self):
        df = pd.DataFrame({
            (PREDICTION_COLUMN_NAME, "a"): [[1, 0, 0], [0, 1, 0], [0, 0, 1]],
            (TARGET_COLUMN_NAME, "a"): [1, 1, 1],
            (TARGET_COLUMN_NAME, "b"): [2, 2, 2],
            (TARGET_COLUMN_NAME, "c"): [3, 3, 3],
        })
        df.columns = pd.MultiIndex.from_tuples(df.columns)

        mapped = map_prediction_to_target(df, PREDICTION_COLUMN_NAME, TARGET_COLUMN_NAME)

        self.assertListEqual(mapped.index.to_list(),[(0, 1), (0, 2), (0, 3), (1, 1), (1, 2), (1, 3), (2, 1), (2, 2), (2, 3)])
        self.assertListEqual(mapped[PREDICTION_COLUMN_NAME].to_list(), [1, 0, 0, 0, 1, 0, 0, 0, 1])

    def test_map_prediction_to_target_open(self):
        df = pd.DataFrame({
            (PREDICTION_COLUMN_NAME, "a"): [[1, 0, 0], [0, 1, 0], [0, 0, 1]],
            (TARGET_COLUMN_NAME, "a"): [2, 2, 2],
            (TARGET_COLUMN_NAME, "b"): [3, 4, 3]
        })
        df.columns = pd.MultiIndex.from_tuples(df.columns)

        mapped = map_prediction_to_target(df, PREDICTION_COLUMN_NAME, TARGET_COLUMN_NAME)

        self.assertListEqual(mapped.index.to_list(), [(0, (-float('inf'), 2)), (0, (2, 3)), (0, (3, float('inf'))), (1, (-float('inf'), 2)), (1, (2, 4)), (1, (4, float('inf'))), (2, (-float('inf'), 2)), (2, (2, 3)), (2, (3, float('inf')))])
        self.assertListEqual(mapped[PREDICTION_COLUMN_NAME].to_list(), [1, 0, 0, 0, 1, 0, 0, 0, 1])

    def test_map_prediction_to_target_closed(self):
        df = pd.DataFrame({
            (PREDICTION_COLUMN_NAME, "a"): [[1, 0, 0], [0, 1, 0], [0, 0, 1]],
            (TARGET_COLUMN_NAME, "a"): [1, 1, 1],
            (TARGET_COLUMN_NAME, "b"): [2, 2, 2],
            (TARGET_COLUMN_NAME, "c"): [3, 3, 3],
            (TARGET_COLUMN_NAME, "d"): [4, 4, 4],
        })
        df.columns = pd.MultiIndex.from_tuples(df.columns)

        mapped = map_prediction_to_target(df, PREDICTION_COLUMN_NAME, TARGET_COLUMN_NAME)
        print(mapped.index.to_list())

        self.assertListEqual(mapped.index.to_list(), [(0, '(1, 2)'), (0, '(2, 3)'), (0, '(3, 4)'), (1, '(1, 2)'), (1, '(2, 3)'), (1, '(3, 4)'), (2, '(1, 2)'), (2, '(2, 3)'), (2, '(3, 4)')])
        self.assertListEqual(mapped[PREDICTION_COLUMN_NAME].to_list(), [1, 0, 0, 0, 1, 0, 0, 0, 1])


import unittest
from unittest import TestCase

import numpy as np
from sklearn.neural_network import MLPClassifier
from sklearn.model_selection import KFold

from pandas_ml_common import naive_splitter, random_splitter, stratified_random_splitter
from pandas_ml_utils import FeaturesAndLabels, SkModel, FittingParameter
from pandas_ml_utils.constants import *
from pandas_ml_utils_test.config import DF_NOTES


class TestModel(TestCase):

    def test_simple_classification_model(self):
        df = DF_NOTES.copy()

        with df.model() as m:
            fit = m.fit(
                SkModel(
                    MLPClassifier(activation='tanh', hidden_layer_sizes=(20, 12), random_state=42, max_iter=2),
                    FeaturesAndLabels(
                        features=["variance", "skewness", "kurtosis", "entropy"],
                        labels=["authentic"],
                        label_type=bool
                    )
                ),
                FittingParameter(stratified_random_splitter())
            )

        print(fit)
        html = fit._repr_html_()

        prediction = df.model.predict(fit.model)
        print(prediction)
        self.assertGreaterEqual(prediction[PREDICTION_COLUMN_NAME].iloc[-1].values, 0.68)

        backtest = df.model.backtest(fit.model)
        self.assertIn(FEATURE_COLUMN_NAME, backtest.df)
        self.assertIn(LABEL_COLUMN_NAME, backtest.df)
        np.testing.assert_array_almost_equal(prediction[PREDICTION_COLUMN_NAME].iloc[-1].values,
                                             backtest.df[PREDICTION_COLUMN_NAME].iloc[-1].values)

        # test multiple samples
        samples = df.model.predict(fit.model, samples=2)
        self.assertIsInstance(samples[PREDICTION_COLUMN_NAME].iloc[-1, 0], list)
        self.assertEqual(2, len(samples[PREDICTION_COLUMN_NAME].iloc[-1, 0]))

    def test_simple_classification_cross_validation(self):
        df = DF_NOTES.copy()

        with df.model() as m:
            fit = m.fit(
                SkModel(
                    MLPClassifier(activation='tanh', hidden_layer_sizes=(20, 12), random_state=42, max_iter=2),
                    FeaturesAndLabels(
                        features=["variance", "skewness", "kurtosis", "entropy"],
                        labels=["authentic"],
                        label_type=bool
                    )
                ),
                FittingParameter(
                    splitter=random_splitter(),
                    cross_validation=KFold(3, random_state=42, shuffle=True)
                )
            )

        print(fit)
        html = fit._repr_html_()

        prediction = df.model.predict(fit.model)
        print(prediction)
        self.assertGreaterEqual(prediction[PREDICTION_COLUMN_NAME].iloc[-1].values, 0.65)

    def test_simple_classification_model_with_all_options(self):
        df = DF_NOTES.copy()

        with df.model() as m:
            fit = m.fit(
                SkModel(
                    MLPClassifier(activation='tanh', hidden_layer_sizes=(20, 12), random_state=42, max_iter=2),
                    FeaturesAndLabels(
                        features=["variance", "skewness", "kurtosis", "entropy"],
                        sample_weights=["variance"],
                        gross_loss=["kurtosis"],
                        targets=["entropy"],
                        labels=["authentic"],
                        label_type=bool
                    )
                ),
                FittingParameter(stratified_random_splitter())
            )

        # should not thro an error
        html = fit._repr_html_()

        # fit resulting columns
        print(fit.test_summary.df)
        self.assertIn(GROSS_LOSS_COLUMN_NAME, fit.training_summary.df)
        self.assertIn(FEATURE_COLUMN_NAME, fit.training_summary.df)
        self.assertIn(LABEL_COLUMN_NAME, fit.training_summary.df)
        self.assertIn(TARGET_COLUMN_NAME, fit.training_summary.df)

        self.assertIn(FEATURE_COLUMN_NAME, fit.test_summary.df)
        self.assertIn(LABEL_COLUMN_NAME, fit.test_summary.df)
        self.assertIn(TARGET_COLUMN_NAME, fit.test_summary.df)

        # prediction resulting columns
        prediction = df.model.predict(fit.model)
        print(prediction)
        self.assertIn(FEATURE_COLUMN_NAME, prediction)
        self.assertIn(TARGET_COLUMN_NAME, prediction)

        # backtest resulting columns
        backtest = df.model.backtest(fit.model)
        print(backtest.df)

        self.assertEqual(len(df), len(backtest.df))
        self.assertIn(FEATURE_COLUMN_NAME, backtest.df)
        self.assertIn(LABEL_COLUMN_NAME, backtest.df)
        self.assertIn(TARGET_COLUMN_NAME, backtest.df)
        self.assertIn(GROSS_LOSS_COLUMN_NAME, backtest.df)


from pandas_ml_utils import Model, pd
import sys


if __name__ == '__main__':
    model_file = sys.argv[1]
    print(f"load model {model_file}")
    model = Model.load(model_file)
    print(model)

    df = pd.DataFrame({
        "variance": [0.40614],
        "skewness": [1.34920],
        "kurtosis": [-1.4501],
        "entropy": [- 0.55949]
    })

    print(df.model.predict(model))




"""Augment pandas DataFrame with methods for machine learning"""
__version__ = '0.2.0'

from pandas_ml_utils.ml.summary import *
from pandas_ml_common import *
from pandas_ml_utils.df_patching.model_patch import DfModelPatch
from pandas_ml_utils.ml.data import *
from pandas_ml_utils.ml.data.extraction import FeaturesAndLabels, PostProcessedFeaturesAndLabels
from pandas_ml_utils.ml.model import *
from pandas_ml_utils.ml.fitting import *

_log = logging.getLogger(__name__)
_log.debug(f"numpy version {np.__version__}")
_log.debug(f"pandas version {pd.__version__}")


setattr(PandasObject, "model", property(lambda self: DfModelPatch(self)))



import logging
from time import perf_counter
from typing import Callable, Tuple, Dict, Union, List

from scipy.stats import randint as sp_randint
from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor
from sklearn.feature_selection import RFECV
from sklearn.model_selection import RandomizedSearchCV, TimeSeriesSplit, KFold, StratifiedKFold

from pandas_ml_common import Typing, naive_splitter, XYWeight
from pandas_ml_common.utils import get_correlation_pairs, merge_kwargs, call_silent, \
    call_callable_dynamic_args
from pandas_ml_utils.ml.data.extraction.features_and_labels_definition import FeaturesAndLabels, \
    PostProcessedFeaturesAndLabels
from pandas_ml_utils.ml.data.reconstruction import assemble_result_frame
from pandas_ml_utils.ml.fitting import Fit, FitException, FittingParameter
from pandas_ml_utils.ml.model import Model as MlModel, SkModel, AutoEncoderModel, SubModelFeature
from pandas_ml_utils.ml.summary import Summary

from ..ml.data.extraction.features_and_labels_extractor import FeaturesWithLabels, FeaturesWithTargets, \
    extract_features, extract_feature_labels_weights
from ..ml.summary.feature_selection_summary import FeatureSelectionSummary

_log = logging.getLogger(__name__)


class DfModelPatch(object):

    def __init__(self, df: Typing.PatchedDataFrame):
        self.df = df
        self._type_mapping = {
            AutoEncoderModel: lambda df, model, **kwargs: df.model.predict(model.as_encoder(), **kwargs),
            MlModel: lambda df, model, **kwargs: df.model.predict(model, **kwargs),
        }

    def feature_selection(self,
                          features_and_labels: FeaturesAndLabels,
                          training_data_splitter: Callable = naive_splitter(0.2),
                          correlated_features_th: float = 0.75,
                          rfecv_splits: int = 4,
                          forest_splits: int = 7,
                          min_features_to_select: int = 1,
                          is_time_series: bool = False,
                          **kwargs):
        assert features_and_labels.label_type in ('regression', 'classification', int, float, bool), \
            "label_type need to be specified: 'regression' | 'classification' !"

        # find best parameters
        with self() as m:
            # extract features and labels
            ext = m.extract(features_and_labels)

            # first perform a correlation analysis and remove correlating features !!!
            if correlated_features_th > 0:
                _, pairs = get_correlation_pairs(ext.features)
                redundant_correlated_features = {i[0]: p for i, p in pairs.items() if p > correlated_features_th}
                _log.warning(f"drop redundant features: {redundant_correlated_features}")

                features_and_labels = PostProcessedFeaturesAndLabels.from_features_and_labels(
                    features_and_labels,
                    feature_post_processor=lambda df: df.drop(redundant_correlated_features.keys(), axis=1)
                )

            # estimate model type and sample properties
            is_classification = 'float' not in (str(features_and_labels.label_type))
            nr_samples = len(self.df)

            if is_classification:
                nr_classes = len(ext.labels.value_counts())
            else:
                nr_classes = max(len(self.df) / 3, 100)

            # estimate grid search parameters
            grid = {
                "estimator__n_estimators": sp_randint(10, 500),
                "estimator__max_depth": [2, None],
                "estimator__min_samples_split": sp_randint(2, nr_samples / nr_classes),
                "estimator__min_samples_leaf": sp_randint(2, nr_samples / nr_classes),
                "estimator__bootstrap": [True, False],
                "estimator__criterion": ["gini", "entropy"] if is_classification else ["mse", "mae"]
            }

            # build model
            cross_validation = TimeSeriesSplit if is_time_series else StratifiedKFold if is_classification else KFold
            estimator = RandomForestClassifier() if is_classification else RandomForestRegressor()
            selector = RFECV(estimator, step=1, cv=cross_validation(rfecv_splits), min_features_to_select=min_features_to_select)
            skm = RandomizedSearchCV(selector, param_distributions=grid, cv=cross_validation(forest_splits), n_jobs=-1)

            # fit model
            fit = m.fit(
                SkModel(skm, features_and_labels=features_and_labels, summary_provider=FeatureSelectionSummary),
                FittingParameter(splitter=training_data_splitter)
            )

        # we hide the loss plot from this summary
        return fit.with_hidden_loss_plot()

    def fit(self,
            model_provider: Callable[[], MlModel],
            fitting_parameter: FittingParameter = FittingParameter(),
            verbose: int = 0,
            callbacks: Union[Callable, List[Callable]] = None,
            fail_silent: bool = False,
            **kwargs
            ) -> Fit:
        df = self.df
        trails = None
        model = model_provider()
        kwargs = merge_kwargs(model.features_and_labels.kwargs, model.kwargs, kwargs)

        def fit_submodel(df, model, **kwargs):
            return model.fit(df, **kwargs)

        typemap_fitting = {SubModelFeature: fit_submodel, **self._type_mapping}
        frames: FeaturesWithLabels = model.features_and_labels(
            df, extract_feature_labels_weights, type_map=typemap_fitting, fitting_parameter=fitting_parameter,
            verbose=verbose, **kwargs
        )

        start_performance_count = perf_counter()
        _log.info("create model")

        df_train_prediction, df_test_prediction = model.fit(
            XYWeight(frames.features, frames.labels, frames.sample_weights),
            fitting_parameter,
            verbose,
            callbacks,
            **kwargs
        )

        _log.info(f"fitting model done in {perf_counter() - start_performance_count: .2f} sec!")

        # assemble result objects
        # get training and test data tuples of the provided frames.
        # NOTE due to event boosting there might be duplicate events in the test data which we need to filter away
        ext_frames = frames.targets, frames.labels, frames.gross_loss, frames.sample_weights, frames.features
        df_train = assemble_result_frame(df_train_prediction, *ext_frames)
        df_test = assemble_result_frame(df_test_prediction[~df_test_prediction.index.duplicated()], *ext_frames)

        # update model properties and return the fit
        model.features_and_labels.set_min_required_samples(frames.features_with_required_samples.min_required_samples)
        model.features_and_labels._kwargs = {k: a for k, a in kwargs.items() if k in model.features_and_labels.kwargs}

        def assemble_fit():
            return Fit(
                model,
                model.summary_provider(df_train, model, is_test=False, **kwargs),
                model.summary_provider(df_test, model, is_test=True, **kwargs),
                trails,
                **kwargs
            )

        return call_silent(assemble_fit, lambda e: FitException(e, model)) if fail_silent else assemble_fit()

    def backtest(self,
                 model: MlModel,
                 summary_provider: Callable[[Typing.PatchedDataFrame], Summary] = None,
                 **kwargs) -> Summary:
        df = self.df
        kwargs = merge_kwargs(model.features_and_labels.kwargs, model.kwargs, kwargs)
        typemap_pred = {SubModelFeature: lambda df, model, **kwargs: model.predict(df, **kwargs), **self._type_mapping}
        frames: FeaturesWithLabels = model.features_and_labels(
            df, extract_feature_labels_weights, type_map=typemap_pred, **kwargs
        )

        predictions = model.predict(frames.features, **kwargs)
        df_backtest = assemble_result_frame(predictions, frames.targets, frames.labels, frames.gross_loss,
                                            frames.sample_weights, frames.features)

        return call_callable_dynamic_args(summary_provider or model.summary_provider, df_backtest, model, **kwargs)

    def predict(self,
                model: MlModel,
                tail: int = None,
                samples: int = 1,
                **kwargs) -> Typing.PatchedDataFrame:
        min_required_samples = model.features_and_labels.min_required_samples
        df = self.df

        if tail is not None:
            if min_required_samples is not None:
                # just use the tail for feature engineering
                df = df[-(abs(tail) + (min_required_samples - 1)):]
            else:
                _log.warning("could not determine the minimum required data from the model")

        kwargs = merge_kwargs(model.features_and_labels.kwargs, model.kwargs, kwargs)
        typemap_pred = {SubModelFeature: lambda df, model, **kwargs: model.predict(df, **kwargs), **self._type_mapping}
        frames: FeaturesWithTargets = model.features_and_labels(
            df, extract_features, type_map=typemap_pred, **kwargs
        )

        predictions = call_callable_dynamic_args(
            model.predict,
            features=frames.features, targets=frames.targets, latent=frames.latent, samples=samples, df=df,
            **kwargs
        )

        return assemble_result_frame(predictions, frames.targets, None, None, None, frames.features)

    def __call__(self, file_name=None):
        from .model_context import ModelContext
        return ModelContext(self.df, file_name=file_name)



import unittest

from sklearn.datasets import make_classification

from pandas_ml_common import stratified_random_splitter
from pandas_ml_utils import pd, FeaturesAndLabels


class TestFeatureSelection(unittest.TestCase):

    def test_feature_selection_classification(self):
        data = make_classification(n_samples=20, n_features=5, n_informative=4, n_redundant=1, n_classes=2)
        df = pd.DataFrame(data[0])
        df["label"] = data[1]

        report = df.model.feature_selection(
            features_and_labels=FeaturesAndLabels(
                features=list(range(5)),
                labels=["label"],
                label_type=int
            ),
            training_data_splitter=stratified_random_splitter(0.5),
            rfecv_splits=2,
            forest_splits=2
        )
        
        print(report)
        # print(report.test_summary.validation_kpis.feature_importance)

    def test_feature_selection_regression(self):
        pass


if __name__ == '__main__':
    unittest.main()



from typing import Tuple, List
from unittest import TestCase

from sklearn.model_selection import KFold

from pandas_ml_common import dummy_splitter
from pandas_ml_utils import Model, FittingParameter
from pandas_ml_utils.constants import *
import pandas as pd
import numpy as np


def create_line_data(n=300, slope=1):
    np.random.seed(32)
    x = np.linspace(0, 1, n)
    y = slope * x + np.random.normal(0, 0.05, n)
    return x, y


def create_sine_data(n=300):
    np.random.seed(32)
    x = np.linspace(0, 1 * 2 * np.pi, n)
    y1 = 3 * np.sin(x)
    y1 = np.concatenate((np.zeros(60), y1 + np.random.normal(0, 0.15 * np.abs(y1), n), np.zeros(60)))
    x = np.concatenate((np.linspace(-3, 0, 60), np.linspace(0, 3 * 2 * np.pi, n),
                        np.linspace(3 * 2 * np.pi, 3 * 2 * np.pi + 3, 60)))
    y2 = 0.1 * x + 1
    y = (y1 + y2) + 2
    return x / x[-1], y


class TestModelAccuracy(TestCase):

    def test_linear_regression(self):
        models = self.provide_linear_regression_model()
        if models is None:
            return

        df = pd.DataFrame(np.array(create_line_data(300)).T, columns=["x", "y"])
        for model, fp in  models:
            with df.model() as m:
                fit = m.fit(model, fp.with_splitter(splitter=dummy_splitter).with_cross_validation(None))
                y = fit.training_summary.df[LABEL_COLUMN_NAME].values
                y_hat = fit.training_summary.df[PREDICTION_COLUMN_NAME].values
                dist = np.sqrt(np.sum((y - y_hat) ** 2))

            print(fp.context, dist)
            losses = list(fit.model._history[("train", 0)].values())

            self.assertLessEqual(dist, 1.22, fp.context)
            if len(losses) > 1:
                self.assertLess(losses[-1], losses[0])

    def test_linear_regression_kfold(self):
        models = self.provide_linear_regression_model()
        if models is None:
            return

        df = pd.DataFrame(np.array(create_line_data(300)).T, columns=["x", "y"])
        for model, fp in models:
            with df.model() as m:
                fit = m.fit(model, fp.with_splitter(splitter=dummy_splitter).with_cross_validation(KFold(4)))
                y = fit.training_summary.df[LABEL_COLUMN_NAME].values
                y_hat = fit.training_summary.df[PREDICTION_COLUMN_NAME].values
                dist = np.sqrt(np.sum((y - y_hat) ** 2))

            print(fp.context, dist)
            self.assertLessEqual(dist, 1.6, fp.context)

    def test_non_linear_regression(self):
        models = self.provide_non_linear_regression_model()
        if models is None:
            return

        df = pd.DataFrame(np.array(create_sine_data(300)).T, columns=["x", "y"])
        for model, fp in models:
            with df.model() as m:
                fit = m.fit(model, fp.with_splitter(splitter=dummy_splitter).with_cross_validation(None))
                y = fit.training_summary.df[LABEL_COLUMN_NAME].values
                y_hat = fit.training_summary.df[PREDICTION_COLUMN_NAME].values
                dist = np.sqrt(np.sum((y - y_hat) ** 2))

            print(fp.context, dist, len(fit.model._history[("train", 0)]))
            self.assertLessEqual(dist, 10, fp.context)

    def provide_linear_regression_model(self) -> List[Tuple[Model, FittingParameter]]:
        return None

    def provide_non_linear_regression_model(self) -> List[Tuple[Model, FittingParameter]]:
        return None




from typing import NamedTuple, Tuple, Callable, Union, List, Dict

from pandas_ml_common import Typing, naive_splitter


class FittingParameter(NamedTuple):
    splitter: Callable[[Typing.PdIndex], Tuple[Typing.PdIndex, Typing.PdIndex]] = naive_splitter()
    filter: Callable[[Typing.PatchedSeries], bool] = None
    cross_validation: Union['BaseCrossValidator', Tuple[int, Callable[[Typing.PdIndex], Tuple[List[int], List[int]]]]] = None
    epochs: int = 1
    batch_size: int = None
    fold_epochs: int = 1
    hyper_parameter_space: Dict = None
    context: str = None

    def with_splitter(self, splitter):
        return FittingParameter(**{**self._asdict(), "splitter": splitter})

    def with_cross_validation(self, cross_validation):
        return FittingParameter(**{**self._asdict(), "cross_validation": cross_validation})



from functools import partial
from unittest import TestCase

import numpy as np
import pandas as pd

from pandas_ml_common import Constant
from pandas_ml_common.decorator import MultiFrameDecorator
from pandas_ml_common.utils.column_lagging_utils import lag_columns
from pandas_ml_utils import FeaturesAndLabels, PostProcessedFeaturesAndLabels
from pandas_ml_utils.ml.data.extraction.features_and_labels_extractor import FeaturesWithLabels
from pandas_ml_utils_test.config import DF_TEST


class TestExtractionOfFeaturesAndLabels(TestCase):

    def test_repr_features_labels(self):
        fnl = FeaturesAndLabels(
            features=[1, "A", lambda x: x * 2],
            labels=[1, "A", lambda x: x * 2],
        )

        # print(repr(fnl).replace('"', "\\\""))
        self.assertEqual(
            "FeaturesAndLabels(	[1, 'A', '            features=[1, \"A\", lambda x: x * 2],\\n'], 	[1, 'A', '            labels=[1, \"A\", lambda x: x * 2],\\n'], 	None, 	None, 	None, 	None, 	None, 	{})",
            repr(fnl)
        )

    def test_repr_post_processed_features_labels(self):
        fnl = PostProcessedFeaturesAndLabels(
            features=[
                "Close",
                lambda df: df["Close"].ta.trix(),
            ],
            feature_post_processor=[
                lambda df: df.ta.rnn(5)
            ],
            labels=[
                Constant(0)
            ],
            targets=[
                lambda df: df["Close"]
            ],
        )

        print(repr(repr(fnl).replace('"', '\"')))

        self.assertEqual(
            'PostProcessedFeaturesAndLabels(\t[\'Close\', \'                lambda df: df["Close"].ta.trix(),\\n\'], \t[\'                lambda df: df.ta.rnn(5)\\n\'], \t[\'Constant(0)\'], \tNone, \tNone, \tNone, \tNone, \tNone, \tNone, \tNone, \t[\'                lambda df: df["Close"]\\n\'], \tNone, \tNone, \t{})',
            repr(fnl)
        )

    def test__extract_label_comprehension(self):
        df = DF_TEST.copy()

        fl: FeaturesWithLabels = df._.extract(
            FeaturesAndLabels(
                features=["Close"],
                labels=[partial(lambda i, df: df["Close"].rename(f"Close_{i}"), i) for i in range(5, 10)]
            )
        )

        #print(labels.tail())
        self.assertListEqual(
            ["Close_5", "Close_6", "Close_7", "Close_8", "Close_9"],
            fl.labels.columns.tolist()
        )

    def test__extract_inf(self):

        df = pd.DataFrame({"a": [1, 2, 4, np.nan, np.inf]})

        with self.assertLogs(level='WARN') as cm:
            fl: FeaturesWithLabels = df._.extract(
                FeaturesAndLabels(
                    features=["a"],
                    labels=["a"]
                )
            )

            self.assertIn("features containing infinit number", cm.output[0])

        self.assertEqual(len(fl.features), 3)

    def test_post_processing(self):
        df = pd.DataFrame({"a": np.arange(20), "b": np.arange(20)})

        fl: FeaturesWithLabels = df._.extract(
            PostProcessedFeaturesAndLabels(
                features=["a"],
                feature_post_processor=lambda df: df * 2,
                labels=["b"],
                labels_post_processor=lambda df: df.loc[df["b"] % 2 == 0]
            )
        )

        self.assertEqual(len(fl.features), 10)
        self.assertEqual(len(fl.labels), 10)
        self.assertEqual(fl.labels.values.sum().item(), 90)

    def test_post_processing_multiindex_row(self):
        df = pd.DataFrame({"a": np.arange(20), "b": np.arange(20)})
        df.index = pd.MultiIndex.from_product([["A", "B"], range(10)])

        fl: FeaturesWithLabels = df._.extract(
            PostProcessedFeaturesAndLabels(
                features=["a"],
                feature_post_processor=lambda df: lag_columns(df, [1, 2]),
                labels=["b"],
                labels_post_processor=lambda df: df * 2,
            )
        )

        self.assertIsInstance(fl.features.index, pd.MultiIndex)
        self.assertIsInstance(fl.labels.index, pd.MultiIndex)

    def test_nested_postprocessors(self):
        df = pd.DataFrame({"a": np.arange(20), "b": np.arange(20), "c": np.arange(20)})

        def outer_postprocessor(df):
            return df[["b"]]

        fl: FeaturesWithLabels = df._.extract(
            PostProcessedFeaturesAndLabels.from_features_and_labels(
                PostProcessedFeaturesAndLabels(
                    features=["a"],
                    feature_post_processor=[lambda df: lag_columns(df, [1, 2])],
                    labels=["b", "c"],
                    labels_post_processor=[lambda df: df + 0.5, lambda df: df + 0.001],
                ),
                labels_post_processor=[outer_postprocessor],
            )

        )

        self.assertEqual((18, 1), fl.labels.shape)
        self.assertEqual(19.501, fl.labels.iloc[-1, -1])

    def test_post_processor_with_multi_frame_decorator(self):
        df = pd.DataFrame({"a": np.arange(20), "b": np.arange(20), "c": np.arange(20)})
        fl: FeaturesWithLabels = df._.extract(
            PostProcessedFeaturesAndLabels(
                features=(["a"], ["b"]),
                feature_post_processor=([lambda df: df - 1], [lambda df: df + 2]),
                labels=["c"]
            )
        )

        self.assertIsInstance(fl.features, MultiFrameDecorator)
        self.assertListEqual([18, 21], fl.features.as_joined_frame().iloc[-1].to_list())



from typing import Tuple, List, Callable

import pandas as pd

from pandas_ml_common import LazyInit


class NbLiveLossPlot(object):

    def __init__(self,
                 figsize: Tuple[int] = None,
                 frequency: int = 10,
                 plot_reconstruction: bool = False,
                 reconstruction_preprocessor: Callable[[pd.DataFrame], pd.DataFrame] = None,
                 backend: str = 'nbAgg'):
        self.frequency = frequency
        self.count = 0
        self.x = []
        self.train_loss = []
        self.val_loss = []
        self.reconstruction_preprocessor = reconstruction_preprocessor if reconstruction_preprocessor is not None else lambda x: x

        # initialize a plot make sure backend is not inline but nbAgg
        import matplotlib.pyplot as plt
        if backend is not None:
            plt.switch_backend(backend.replace('notebook', 'nbAgg'))

        self.fig, self.ax = plt.subplots(2 if plot_reconstruction else 1, 2, figsize=figsize)
        self.ax = self.ax.flatten()
        self.fig.show()

        # from IPython.display import display
        # display(self.fig)

    def __call__(self, epoch, fold, fold_epoch, loss, val_loss, y_train, y_test: List, y_hat_train: LazyInit, y_hat_test: List[LazyInit]):
        # print(epoch, fold, fold_epoch, loss, val_loss)
        self.count += 1
        self.x.append(self.x[-1] + 1 if len(self.x) > 0 else 0)
        self.train_loss.append(loss)
        self.val_loss.append(val_loss)

        if self.count % self.frequency == 0:
            for x in self.ax: x.clear()
            self.ax[0].plot(self.x, self.train_loss)
            self.ax[1].plot(self.x, self.val_loss)
            self.ax[0].title.set_text('Training Loss')
            self.ax[1].title.set_text('Test Loss')

            if len(self.ax) > 2:
                self.ax[2].plot(self.reconstruction_preprocessor(y_train))
                self.ax[2].plot(self.reconstruction_preprocessor(y_hat_train()))
                if len(y_test) > 0:
                    self.ax[3].plot(self.reconstruction_preprocessor(y_test[0]))
                    self.ax[3].plot(self.reconstruction_preprocessor(y_hat_test[0]()))

                self.ax[2].title.set_text('Training Reconstruction')
                self.ax[3].title.set_text('Test Reconstruction')

            self.fig.canvas.draw()




from unittest import TestCase

from pandas_ml_utils.df_patching.model_patch import DfModelPatch
from pandas_ml_utils_test.config import DF_TEST


class TestDfModelExtension(TestCase):

    def test__property(self):
        self.assertIsInstance(DF_TEST.model, DfModelPatch)




from collections import defaultdict
from collections import defaultdict
from itertools import chain
from typing import Callable
import logging
import numpy as np
import pandas as pd
from sklearn import metrics
from sklearn.metrics import roc_curve, auc, mean_squared_error

from pandas_ml_common.utils import get_correlation_pairs, unique_level_rows, call_callable_dynamic_args
from pandas_ml_common.utils.numpy_utils import clean_one_hot_classification
from pandas_ml_utils.constants import *

_log = logging.getLogger(__name__)


def plot_true_pred_scatter(df, figsize=(6, 6), alpha=0.1, **kwargs):
    import matplotlib.pyplot as plt

    x = df[PREDICTION_COLUMN_NAME].values
    y = df[LABEL_COLUMN_NAME].values
    mn = min(x.min(), y.min())
    mx = max(x.max(), y.max())

    fig, axis = plt.subplots(figsize=figsize)
    plt.axes(aspect='equal')
    plt.xlabel("Predicted")
    plt.ylabel("True")

    plt.scatter(x, y, alpha=alpha)
    plt.xlim([mn, mx])
    plt.ylim([mn, mx])
    plt.plot([mn, mx], [mn, mx])

    return fig


def plot_receiver_operating_characteristic(df, figsize=(6, 6), **kwargs):
    import matplotlib.pyplot as plt

    # get true and prediction data. It needs to be a one hot encoded 2D array [samples, class] where nr_classes >= 2
    tv, pv = clean_one_hot_classification(df[LABEL_COLUMN_NAME]._.values, df[PREDICTION_COLUMN_NAME]._.values)

    # calculate legends
    legend = [(col[1] if isinstance(col, tuple) else col) for col in df[LABEL_COLUMN_NAME].columns.tolist()]
    if tv.shape[1] > len(legend):
        legend = list(range(tv.shape[1]))

    # Compute ROC curve and ROC area for each class
    fpr = dict()
    tpr = dict()
    roc_auc = dict()

    for i in range(tv.shape[1]):
        fpr[i], tpr[i], _ = roc_curve(tv[:, i], pv[:, i])
        roc_auc[i] = auc(fpr[i], tpr[i])

    # plot ROC curves
    fig, axis = plt.subplots(figsize=figsize)

    for i in fpr.keys():
        plt.plot(fpr[i], tpr[i], label=f"{legend[i]} auc:{roc_auc[i] * 100:.2f}")

    plt.plot([0, 1], [0, 1], color='navy', linestyle='--')
    plt.xlim([0.0, 1.0])
    plt.ylim([0.0, 1.05])
    plt.xlabel('False Positive Rate')
    plt.ylabel('True Positive Rate')
    plt.title('Receiver Operating Characteristic')
    plt.legend(loc="lower right")
    return fig


def plot_confusion_matrix(df, figsize=(6, 6), **kwargs):
    from mlxtend.plotting import plot_confusion_matrix
    from mlxtend.evaluate import confusion_matrix

    # get true and prediction data. It needs to be a one hot encoded 2D array [samples, class] where nr_classes >= 2
    tv, pv = clean_one_hot_classification(df[LABEL_COLUMN_NAME]._.values, df[PREDICTION_COLUMN_NAME]._.values)

    # confusion matrix needs integer encoding
    tv = np.apply_along_axis(np.argmax, 1, tv)
    pv = np.apply_along_axis(np.argmax, 1, pv)

    # plot the confusion matrix
    cm = confusion_matrix(tv, pv, binary=tv.max() < 2)
    fig, ax = plot_confusion_matrix(cm, figsize=figsize)

    return fig


def plot_feature_correlation(df, model=None, cmap='bwr', width=15, **kwargs):
    import matplotlib.pyplot as plt

    # extract features or use whole data frame
    if FEATURE_COLUMN_NAME in df:
        df = df[FEATURE_COLUMN_NAME]

    # cet correlations
    corr, sorted_corr = get_correlation_pairs(df)

    # generate plot
    fig, (ax, ax2) = plt.subplots(1, 2, figsize=(width, width / 1.33), gridspec_kw={'width_ratios': [3, 1]})
    ax.pcolor(corr, cmap=cmap)
    for i, j in np.ndindex(corr.shape):
        ax.text(j + 0.5, i + 0.5, f'{corr.iloc[i, j]:.2f}', ha="center", va="center", color="w")

    ax.xaxis.tick_top()
    ax.set_yticks(np.arange(0.5, len(corr), 1))
    ax.set_yticklabels(corr.index)
    ax.set_xticks(np.arange(0.5, len(corr), 1))
    ax.set_xticklabels(corr.columns, rotation=60, ha="left", rotation_mode="anchor")

    ax2.pcolor(sorted_corr.values.reshape(-1, 1), cmap=cmap)
    for i, v in enumerate(sorted_corr.values):
        ax2.text(0.5, i + 0.5, f'{v:.2f}', ha="center", va="center", color="w")

    ax2.set_xticks([0, 1])
    ax2.set_xticklabels([])
    ax2.yaxis.tick_right()
    ax2.set_yticks(np.arange(0.5, len(sorted_corr), 1))
    ax2.set_yticklabels([f'{a} / {b}' for a, b in sorted_corr.index])

    return fig


def plot_feature_importance(df, model, **kwargs):
    import matplotlib.pyplot as plt

    if FEATURE_COLUMN_NAME not in df:
        return f"DataFrame does not contain {FEATURE_COLUMN_NAME}!"

    mutation_steps = len(df) // 2
    dff = df[FEATURE_COLUMN_NAME]
    dfl = df[LABEL_COLUMN_NAME]
    dfw = df[SAMPLE_WEIGHTS_COLUMN_NAME] if SAMPLE_WEIGHTS_COLUMN_NAME in df else None
    loss_star = model.calculate_loss(None, dff, dfl, dfw)
    permuted_feature_loss = defaultdict(lambda: [])

    for col in dff.columns:
        x = dff.copy()
        x[col] = np.roll(x[col], mutation_steps)
        loss = model.calculate_loss(None, x, dfl, dfw)
        permuted_feature_loss[col].append(loss - loss_star)

    def sort_by_mean_error(item):
        return np.array(item[1]).mean()

    # more important features should cause a higher loss
    permuted_feature_loss = {k: v for k, v in sorted(permuted_feature_loss.items(), key=sort_by_mean_error, reverse=True)}

    fig, ax = plt.subplots(1, 1, figsize=(20, 6))
    pd.DataFrame(permuted_feature_loss).plot.bar(ax=ax)
    ax.set_title('Featureimportance')
    return fig


def df_tail(df, model, **kwargs):
    return df.tail()


def df_regression_scores(df, model, **kwargs):
    rm = metrics._regression
    ALL = [
        "r2_score",
        "explained_variance_score",
        "mean_gamma_deviance",
        "max_error",
        "mean_absolute_error",
        "mean_squared_error",
        "mean_squared_log_error",
        "median_absolute_error",
        "mean_tweedie_deviance",
        "mean_poisson_deviance",
    ]

    def score_regression(df):
        y_true = df[LABEL_COLUMN_NAME]._.values
        y_pred = df[PREDICTION_COLUMN_NAME]._.values
        sample_weights = df[SAMPLE_WEIGHTS_COLUMN_NAME] if SAMPLE_WEIGHTS_COLUMN_NAME in df else None
        scores = defaultdict(lambda: [])

        for scorer in ALL:
            try:
                score = call_callable_dynamic_args(rm.__dict__[scorer], y_true, y_pred, sample_weight=sample_weights)
                scores[scorer].append(score)
            except Exception as e:
                _log.warning(f"{scorer} failed: {str(e)[:160]}")
                scores[scorer].append(np.nan)

        return pd.DataFrame(scores)

    fig_df = score_regression(df) if not isinstance(df.index, pd.MultiIndex) else pd.concat(
        [score_regression(df.loc[group]).add_multi_index(group, inplace=True, axis=0) for group in unique_level_rows(df)],
        axis=0
    )

    if kwargs.get("no_style", False):
        return fig_df.T

    try:
        # we try to apply styling for relative indicators if matplotlib is installed
        import matplotlib.pyplot as plt
        from matplotlib.colors import to_hex
        cmap = plt.get_cmap('RdYlGn')

        def color_scaler(x, min, max, sign=1):
            out = (0, 1) if sign > 0 else (1, 0)
            try:
                col_val = np.interp(x, (min, max), out)
            except:
                col_val = out[0]

            return to_hex(cmap(col_val)) + "95"

        return fig_df.T.style\
            .apply(lambda x: [f"background-color: {color_scaler(x.item(), -1, 1)}"],
                   subset=pd.IndexSlice["r2_score", :]) \
            .apply(lambda x: [f"background-color: {color_scaler(x.item(), -1, 1)}"],
                   subset=pd.IndexSlice["explained_variance_score", :]) \
            .apply(lambda x: [f"background-color: {color_scaler(x.item(), 2, 0, -1)}"],
                   subset=pd.IndexSlice["mean_gamma_deviance", :]) \
            .set_precision(2)

    except:
        return fig_df.T


def df_classification_scores(df, model, **kwargs):
    class_scores = {  # [worst, best]
        "balanced_accuracy_score":  [0, 1],
        "cohen_kappa_score":    [0, 1],
        "matthews_corrcoef":    [0, 1],
        "f1_score": [0, 1],
        "jaccard_score":    [0, 1],
        "precision_score":  [0, 1],
        "recall_score": [0, 1],
        "accuracy_score":   [0, 1],
    }
    losses = {
        "roc_auc_score": [0.5, 1],
        "zero_one_loss": None,
        "hinge_loss": None,
        "log_loss": None,
        "hamming_loss": None,
    }

    def score_classification(df):
        y_true = df[LABEL_COLUMN_NAME]._.values
        y_pred = df[PREDICTION_COLUMN_NAME]._.values
        sample_weights = df[SAMPLE_WEIGHTS_COLUMN_NAME] if SAMPLE_WEIGHTS_COLUMN_NAME in df else None
        scores = defaultdict(lambda: [])

        y_true_class = np.argmax(y_true, axis=1) if y_true.ndim > 1 and y_true.shape[1] > 1 else y_true
        y_pred_class = np.argmax(y_pred, axis=1) if y_pred.ndim > 1 and y_pred.shape[1] > 1 else y_pred > 0.5

        for scorer in class_scores.keys():
            try:
                score = call_callable_dynamic_args(metrics.__dict__[scorer], y_true_class, y_pred_class, sample_weight=sample_weights)
                if scorer == 'log_loss' and y_pred.ndim > 1:
                    score /= np.log(y_pred.shape[1])

                scores[scorer].append(score)
            except Exception as e:
                _log.warning(f"{scorer} failed: {str(e)[:160]}")
                scores[scorer].append(np.nan)

        for scorer in losses.keys():
            try:
                score = call_callable_dynamic_args(metrics.__dict__[scorer], y_true_class, y_pred, sample_weight=sample_weights)
                scores[scorer].append(score)
            except Exception as e:
                _log.warning(f"{scorer} failed: {str(e)[:160]}")
                scores[scorer].append(np.nan)

        return pd.DataFrame(scores)

    fig_df = score_classification(df) if not isinstance(df.index, pd.MultiIndex) else pd.concat(
        [score_classification(df.loc[group]).add_multi_index(group, inplace=True, axis=0) for group in unique_level_rows(df)],
        axis=0
    )

    if kwargs.get("no_style", False):
        return fig_df.T

    try:
        # we try to apply styling for relative indicators if matplotlib is installed
        import matplotlib.pyplot as plt
        from matplotlib.colors import to_hex
        cmap = plt.get_cmap('RdYlGn')

        def color_scaler(x, min, max):
            col_val = np.interp(x, (min, max), (0, 1))
            return to_hex(cmap(col_val)) + "95"

        styled = fig_df.T.style.set_precision(2)
        for scorer, domain in chain(class_scores.items(), losses.items()):
            if domain is None:
                continue
            styled = styled.apply(lambda x, d=domain: [f"background-color: {color_scaler(x.item(), *d)}"],
                                  subset=pd.IndexSlice[scorer, :])

        return styled
    except Exception as e:
        return fig_df.T





from unittest import TestCase

import numpy as np
from sklearn.metrics import confusion_matrix

from pandas_ml_utils import FeaturesAndLabels, LambdaModel
from pandas_ml_utils.constants import *
from pandas_ml_utils_test.config import DF_NOTES


class TestLambdaModel(TestCase):

    def test_lambda_model_classifier(self):
        df = DF_NOTES.copy()

        with df.model() as m:
            fit = m.fit(
                LambdaModel(
                    lambda f: np.abs(f.values) > 3.5,
                    FeaturesAndLabels(features=["variance"], labels=["authentic"])
                )
            )

        cm = confusion_matrix(
            fit.test_summary.df[LABEL_COLUMN_NAME].values,
            fit.test_summary.df[PREDICTION_COLUMN_NAME].values
        )

        tn, fp, fn, tp = cm.ravel()
        self.assertEqual(89, tp)



from __future__ import annotations

import logging
from typing import Callable, Union

import numpy as np
import pandas as pd

from pandas_ml_common import Typing
from pandas_ml_common.utils import to_pandas, call_callable_dynamic_args
from pandas_ml_utils.ml.data.extraction import FeaturesAndLabels
from pandas_ml_utils.ml.summary import Summary
from .base_model import Model

_log = logging.getLogger(__name__)


class LambdaModel(Model):

    def __init__(self,
                 model: Callable[[pd.DataFrame], Union[np.ndarray, pd.DataFrame]],
                 features_and_labels: FeaturesAndLabels,
                 summary_provider: Callable[[Typing.PatchedDataFrame], Summary] = Summary,
                 **kwargs):
        super().__init__(features_and_labels, summary_provider, **kwargs)
        self.model = model

    def fit_batch(self, x: pd.DataFrame, y: pd.DataFrame, w: pd.DataFrame, fold: int, **kwargs):
        pass

    def calculate_loss(self, fold: int, x: pd.DataFrame, y_true: pd.DataFrame, weight: pd.DataFrame) -> float:
        return 0

    def predict(self, features: pd.DataFrame, targets: pd.DataFrame = None, latent: pd.DataFrame = None, samples=1, **kwargs) -> Typing.PatchedDataFrame:
        pred = call_callable_dynamic_args(self.model, features, targets=targets, **self.kwargs)

        if isinstance(pred, pd.DataFrame):
            pred.columns = self._labels_columns
            return pred
        else:
            return to_pandas(pred, features.index, self._labels_columns)



from unittest import TestCase
import pandas as pd
from pandas_ml_utils import ClassificationSummary, RegressionSummary
import numpy as np
from pandas_ml_utils.constants import *


class TestSummary(TestCase):

    def test_classification_summary(self):
        df = pd.DataFrame({"a": np.hstack([np.zeros(5), np.ones(5)]), "b": np.random.random(10)})
        df.columns = pd.MultiIndex.from_product([[LABEL_COLUMN_NAME, PREDICTION_COLUMN_NAME], ["a"]])

        html = ClassificationSummary(df, None)._repr_html_()

        # render html http://htmledit.squarefree.com/
        print(html)

        self.assertIn("data:image/png;base64", html)

    def test_regression_summary(self):
        df = pd.DataFrame({"a": np.random.random(10), "b": np.random.random(10)})
        df.columns = pd.MultiIndex.from_product([[LABEL_COLUMN_NAME, PREDICTION_COLUMN_NAME], ["a"]])

        html = RegressionSummary(df, None)._repr_html_()

        # render html http://htmledit.squarefree.com/
        print(html)

        self.assertIn("data:image/png;base64", html)

